from __future__ import annotations

import logging
from typing import Optional, List, Any, Dict, overload, Iterable
from uuid import UUID

import sqlalchemy.orm as orm

import pixeltable
import pixeltable.type_system as ts
from pixeltable import exceptions as excs
from pixeltable.env import Env
from .catalog import Catalog
from .table import Table
from .table_version import TableVersion
from .table_version_path import TableVersionPath

_logger = logging.getLogger('pixeltable')

class InsertableTable(Table):
    """A `Table` that allows inserting and deleting rows."""
    def __init__(self, dir_id: UUID, tbl_version: TableVersion):
        tbl_version_path = TableVersionPath(tbl_version)
        super().__init__(tbl_version.id, dir_id, tbl_version.name, tbl_version_path)

    @classmethod
    def display_name(cls) -> str:
        return 'table'

    # MODULE-LOCAL, NOT PUBLIC
    @classmethod
    def create(
            cls, dir_id: UUID, name: str, schema: Dict[str, ts.ColumnType], primary_key: List[str],
            num_retained_versions: int, comment: str
    ) -> InsertableTable:
        columns = cls._create_columns(schema)
        cls._verify_schema(columns)
        column_names = [col.name for col in columns]
        for pk_col in primary_key:
            if pk_col not in column_names:
                raise excs.Error(f'Primary key column {pk_col} not found in table schema')
            col = columns[column_names.index(pk_col)]
            if col.col_type.nullable:
                raise excs.Error(f'Primary key column {pk_col} cannot be nullable')
            col.primary_key = True

        with orm.Session(Env.get().engine, future=True) as session:
            _, tbl_version = TableVersion.create(session, dir_id, name, columns, num_retained_versions, comment)
            tbl = cls(dir_id, tbl_version)
            session.commit()
            cat = Catalog.get()
            cat.tbl_dependents[tbl._id] = []
            cat.tbls[tbl._id] = tbl

            _logger.info(f'Created table `{name}`, id={tbl_version.id}')
            print(f'Created table `{name}`.')
            return tbl

    @overload
    def insert(self, rows: Iterable[Dict[str, Any]], /, print_stats: bool = False, fail_on_exception: bool = True): ...

    @overload
    def insert(self, print_stats: bool = False, fail_on_exception: bool = True, **kwargs: Any): ...

    def insert(self, *args, **kwargs) -> Table.UpdateStatus:
        """Insert rows into table.

        To insert multiple rows at a time:

        ``insert(rows: List[Dict[str, Any]], print_stats: bool = False, fail_on_exception: bool = True)``

        To insert just a single row, you can use the more convenient syntax:
        ``insert(print_stats: bool = False, fail_on_exception: bool = True, **kwargs: Any)``

        Args:
            rows: (if inserting multiple rows) A list of rows to insert, each of which is a dictionary mapping column
                names to values.
            kwargs: (if inserting a single row) keyword-argument pairs representing column names and values.
            print_stats: If ``True``, print statistics about the cost of computed columns.
            fail_on_exception:
                Determines how exceptions in computed columns and invalid media files (e.g., corrupt images)
                are handled.
                If ``False``, store error information (accessible as column properties 'errortype' and 'errormsg')
                for those cases, but continue inserting rows.
                If ``True``, raise an exception that aborts the insert.

        Returns:
            execution status

        Raises:
            Error: if a row does not match the table schema or contains values for computed columns

        Examples:
            Insert two rows into a table with three int columns ``a``, ``b``, and ``c``. Column ``c`` is nullable.

            >>> tbl.insert([{'a': 1, 'b': 1, 'c': 1}, {'a': 2, 'b': 2}])

            Insert a single row into a table with three int columns ``a``, ``b``, and ``c``.

            >>> tbl.insert(a=1, b=1, c=1)
        """
        print_stats = kwargs.pop('print_stats', False)
        fail_on_exception = kwargs.pop('fail_on_exception', True)
        if len(args) > 0:
            # There's a positional argument; this means `rows` is expressed as a
            # list of dicts (multi-insert)
            rows = list(args[0])
        else:
            # No positional argument; this means we're inserting a single row
            # using kwargs syntax
            rows = [kwargs]

        if not isinstance(rows, list):
            raise excs.Error('rows must be a list of dictionaries')
        if len(rows) == 0:
            raise excs.Error('rows must not be empty')
        for row in rows:
            if not isinstance(row, dict):
                raise excs.Error('rows must be a list of dictionaries')
        self._validate_input_rows(rows)
        result = self.tbl_version.insert(rows, print_stats=print_stats, fail_on_exception=fail_on_exception)

        if result.num_excs == 0:
            cols_with_excs_str = ''
        else:
            cols_with_excs_str = \
                f' across {len(result.cols_with_excs)} column{"" if len(result.cols_with_excs) == 1 else "s"}'
            cols_with_excs_str += f' ({", ".join(result.cols_with_excs)})'
        msg = (
            f'Inserted {result.num_rows} row{"" if result.num_rows == 1 else "s"} '
            f'with {result.num_excs} error{"" if result.num_excs == 1 else "s"}{cols_with_excs_str}.'
        )
        print(msg)
        _logger.info(f'InsertableTable {self._name}: {msg}')
        return result

    def _validate_input_rows(self, rows: List[Dict[str, Any]]) -> None:
        """Verify that the input rows match the table schema"""
        valid_col_names = set(self.column_names())
        reqd_col_names = set(self.tbl_version_path.tbl_version.get_required_col_names())
        computed_col_names = set(self.tbl_version_path.tbl_version.get_computed_col_names())
        for row in rows:
            assert isinstance(row, dict)
            col_names = set(row.keys())
            if len(reqd_col_names - col_names) > 0:
                raise excs.Error(f'Missing required column(s) ({", ".join(reqd_col_names - col_names)}) in row {row}')

            for col_name, val in row.items():
                if col_name not in valid_col_names:
                    raise excs.Error(f'Unknown column name {col_name} in row {row}')
                if col_name in computed_col_names:
                    raise excs.Error(f'Value for computed column {col_name} in row {row}')

                # validate data
                col = self.tbl_version_path.get_column(col_name)
                try:
                    # basic sanity checks here
                    checked_val = col.col_type.create_literal(val)
                    row[col_name] = checked_val
                except TypeError as e:
                    msg = str(e)
                    raise excs.Error(f'Error in column {col.name}: {msg[0].lower() + msg[1:]}\nRow: {row}')

    def delete(self, where: Optional['pixeltable.exprs.Predicate'] = None) -> Table.UpdateStatus:
        """Delete rows in this table.

        Args:
            where: a Predicate to filter rows to delete.

        Examples:
            Delete all rows in a table:

            >>> tbl.delete()

            Delete all rows in a table where column `a` is greater than 5:

            >>> tbl.delete(tbl.a > 5)
        """
        from pixeltable.exprs import Predicate
        from pixeltable.plan import Planner
        if where is not None:
            if not isinstance(where, Predicate):
                raise excs.Error(f"'where' argument must be a Predicate, got {type(where)}")
            analysis_info = Planner.analyze(self.tbl_version, where)
            if analysis_info.similarity_clause is not None:
                raise excs.Error('nearest() cannot be used with delete()')
            # for now we require that the updated rows can be identified via SQL, rather than via a Python filter
            if analysis_info.filter is not None:
                raise excs.Error(f'Filter {analysis_info.filter} not expressible in SQL')

        return self.tbl_version.delete(where)



from __future__ import annotations
from typing import Optional, List, Dict

import sqlalchemy as sql

from .predicate import Predicate
from .expr import Expr
from .data_row import DataRow
from .row_builder import RowBuilder
import pixeltable.catalog as catalog


class IsNull(Predicate):
    def __init__(self, e: Expr):
        super().__init__()
        self.components = [e]
        self.id = self._create_id()

    def __str__(self) -> str:
        return f'{str(self.components[0])} == None'

    def _equals(self, other: IsNull) -> bool:
        return True

    def sql_expr(self) -> Optional[sql.ClauseElement]:
        e = self.components[0].sql_expr()
        if e is None:
            return None
        return e == None

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        data_row[self.slot_idx] = data_row[self.components[0].slot_idx] is None

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr]) -> Expr:
        assert len(components) == 1
        return cls(components[0])




from __future__ import annotations
from typing import Optional, List, Any, Dict, Tuple, Callable
import operator

import sqlalchemy as sql

from .expr import Expr
from .globals import LogicalOperator
from .predicate import Predicate
from .data_row import DataRow
from .row_builder import RowBuilder
import pixeltable.catalog as catalog


class CompoundPredicate(Predicate):
    def __init__(self, operator: LogicalOperator, operands: List[Predicate]):
        super().__init__()
        self.operator = operator
        # operands are stored in self.components
        if self.operator == LogicalOperator.NOT:
            assert len(operands) == 1
            self.components = operands
        else:
            assert len(operands) > 1
            self.operands: List[Predicate] = []
            for operand in operands:
                self._merge_operand(operand)

        self.id = self._create_id()

    def __str__(self) -> str:
        if self.operator == LogicalOperator.NOT:
            return f'~({self.components[0]})'
        return f' {self.operator} '.join([f'({e})' for e in self.components])

    @classmethod
    def make_conjunction(cls, operands: List[Predicate]) -> Optional[Predicate]:
        if len(operands) == 0:
            return None
        if len(operands) == 1:
            return operands[0]
        return CompoundPredicate(LogicalOperator.AND, operands)

    def _merge_operand(self, operand: Predicate) -> None:
        """
        Merge this operand, if possible, otherwise simply record it.
        """
        if isinstance(operand, CompoundPredicate) and operand.operator == self.operator:
            # this can be merged
            for child_op in operand.components:
                self._merge_operand(child_op)
        else:
            self.components.append(operand)

    def _equals(self, other: CompoundPredicate) -> bool:
        return self.operator == other.operator

    def _id_attrs(self) -> List[Tuple[str, Any]]:
        return super()._id_attrs() + [('operator', self.operator.value)]

    def split_conjuncts(
            self, condition: Callable[[Predicate], bool]) -> Tuple[List[Predicate], Optional[Predicate]]:
        if self.operator == LogicalOperator.OR or self.operator == LogicalOperator.NOT:
            return super().split_conjuncts(condition)
        matches = [op for op in self.components if condition(op)]
        non_matches = [op for op in self.components if not condition(op)]
        return (matches, self.make_conjunction(non_matches))

    def sql_expr(self) -> Optional[sql.ClauseElement]:
        sql_exprs = [op.sql_expr() for op in self.components]
        if any(e is None for e in sql_exprs):
            return None
        if self.operator == LogicalOperator.NOT:
            assert len(sql_exprs) == 1
            return sql.not_(sql_exprs[0])
        assert len(sql_exprs) > 1
        operator = sql.and_ if self.operator == LogicalOperator.AND else sql.or_
        combined = operator(*sql_exprs)
        return combined

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        if self.operator == LogicalOperator.NOT:
            data_row[self.slot_idx] = not data_row[self.components[0].slot_idx]
        else:
            val = True if self.operator == LogicalOperator.AND else False
            op_function = operator.and_ if self.operator == LogicalOperator.AND else operator.or_
            for op in self.components:
                val = op_function(val, data_row[op.slot_idx])
            data_row[self.slot_idx] = val

    def _as_dict(self) -> Dict:
        return {'operator': self.operator.value, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr]) -> Expr:
        assert 'operator' in d
        return cls(LogicalOperator(d['operator']), components)




from typing import List, Any
from types import TracebackType
from dataclasses import dataclass


class Error(Exception):
    pass


@dataclass
class ExprEvalError(Exception):
    expr: Any  # exprs.Expr, but we're not importing pixeltable.exprs to avoid circular imports
    expr_msg: str
    exc: Exception
    exc_tb: TracebackType
    input_vals: List[Any]
    row_num: int


import pytest

import pixeltable as pxt
import pixeltable.exceptions as excs


class TestClient:
    def test_list_functions(self, init_env) -> None:
        cl = pxt.Client()
        _ = cl.list_functions()
        print(_)

    def test_drop_table(self, test_tbl: pxt.Table) -> None:
        cl = pxt.Client()
        t = cl.get_table('test_tbl')
        cl.drop_table('test_tbl')
        with pytest.raises(excs.Error):
            _ = cl.get_table('test_tbl')
        with pytest.raises(excs.Error):
            _ = t.show(1)




import itertools
import json
import re
from typing import Optional, Set, List

import pytest

import pixeltable as pxt
from pixeltable.iterators.document import DocumentSplitter
from pixeltable.tests.utils import get_documents, get_video_files, get_audio_files, get_image_files
from pixeltable.type_system import DocumentType


class TestDocument:
    def valid_doc_paths(self) -> List[str]:
        return get_documents()

    def invalid_doc_paths(self) -> List[str]:
        return [get_video_files()[0], get_audio_files()[0], get_image_files()[0]]

    def test_insert(self, test_client: pxt.Client) -> None:
        file_paths = self.valid_doc_paths()
        cl = test_client
        doc_t = cl.create_table('docs', {'doc': DocumentType()})
        status = doc_t.insert({'doc': p} for p in file_paths)
        assert status.num_rows == len(file_paths)
        assert status.num_excs == 0
        stored_paths = doc_t.select(output=doc_t.doc.localpath).collect()['output']
        assert set(stored_paths) == set(file_paths)

        file_paths = self.invalid_doc_paths()
        status = doc_t.insert(({'doc': p} for p in file_paths), fail_on_exception=False)
        assert status.num_rows == len(file_paths)
        assert status.num_excs == len(file_paths)

    def test_doc_splitter(self, test_client: pxt.Client) -> None:
        file_paths = self.valid_doc_paths()
        cl = test_client
        doc_t = cl.create_table('docs', {'doc': DocumentType()})
        status = doc_t.insert({'doc': p} for p in file_paths)
        assert status.num_excs == 0

        def normalize(s: str) -> str:
            # remove whitespace
            res = re.sub(r'\s+', '', s)
            # remove non-ascii
            res = res.encode('ascii', 'ignore').decode()
            return res

        # run all combinations of (heading, paragraph, sentence) x (token_limit, char_limit, None)
        # and make sure they extract the same text in aggregate
        all_text_reference: Optional[str] = None  # all text as a single string; normalized
        headings_reference: Set[str] = {}  # headings metadata as a json-serialized string
        for sep1 in ['heading', 'paragraph', 'sentence']:
            for sep2 in [None, 'token_limit', 'char_limit']:
                chunk_limits = [10, 20, 100] if sep2 is not None else [None]
                for limit in chunk_limits:
                    iterator_args = {
                        'document': doc_t.doc,
                        'separators': sep1 + (',' + sep2 if sep2 is not None else ''),
                        'metadata': 'title,headings,sourceline'
                    }
                    if sep2 is not None:
                        iterator_args['limit'] = limit
                        iterator_args['overlap'] = 0
                    chunks_t = cl.create_view(
                        f'chunks', doc_t, iterator_class=DocumentSplitter, iterator_args=iterator_args)
                    res = list(chunks_t.order_by(chunks_t.doc, chunks_t.pos).collect())

                    if all_text_reference is None:
                        all_text_reference = normalize(''.join([r['text'] for r in res]))
                        headings_reference = set(json.dumps(r['headings']) for r in res)
                    else:
                        all_text = normalize(''.join([r['text'] for r in res]))
                        headings = set(json.dumps(r['headings']) for r in res)

                        # for debugging
                        first_diff_index = next(
                            (i for i, (c1, c2) in enumerate(zip(all_text, all_text_reference)) if c1 != c2),
                            len(all_text) if len(all_text) != len(all_text_reference) else None)
                        if first_diff_index is not None:
                            a = all_text[max(0, first_diff_index - 10):first_diff_index + 10]
                            b = all_text_reference[max(0, first_diff_index - 10):first_diff_index + 10]

                        assert all_text == all_text_reference, f'{sep1}, {sep2}, {limit}'
                        assert headings == headings_reference, f'{sep1}, {sep2}, {limit}'
                        # TODO: verify chunk limit
                    cl.drop_table('chunks')

    def test_doc_splitter_headings(self, test_client: pxt.Client) -> None:
        file_paths = self.valid_doc_paths()
        cl = test_client
        doc_t = cl.create_table('docs', {'doc': DocumentType()})
        status = doc_t.insert({'doc': p} for p in file_paths)
        assert status.num_excs == 0

        # verify that only the requested metadata is present in the view
        md_elements = ['title', 'headings', 'sourceline']
        md_tuples = list(itertools.chain.from_iterable(itertools.combinations(md_elements, i) for i in range(len(md_elements) + 1)))
        _ = [','.join(t) for t in md_tuples]
        for md_str in [','.join(t) for t in md_tuples]:
            iterator_args = {
                'document': doc_t.doc,
                'separators': 'sentence',
                'metadata': md_str
            }
            chunks_t = cl.create_view(
                f'chunks', doc_t, iterator_class=DocumentSplitter, iterator_args=iterator_args)
            res = chunks_t.order_by(chunks_t.doc, chunks_t.pos).collect()
            requested_md_elements = set(md_str.split(','))
            for md_element in md_elements:
                if md_element in requested_md_elements:
                    _ = res[md_element]
                else:
                    with pytest.raises(pxt.Error):
                        _ = res[md_element]
            cl.drop_table('chunks')



from __future__ import annotations

import copy
import logging
from typing import Optional, List, Dict, Type
from uuid import UUID

import sqlalchemy.orm as orm

from pixeltable import exceptions as excs
from pixeltable.env import Env
from pixeltable.metadata import schema
from .dir import Dir
from .path import Path
from .schema_object import SchemaObject

_logger = logging.getLogger('pixeltable')

class PathDict:
    """Keep track of all paths in a Db instance"""
    def __init__(self):
        self.dir_contents: Dict[UUID, Dict[str, SchemaObject]] = {}
        self.schema_objs: Dict[UUID, SchemaObject] = {}

        # load dirs
        with orm.Session(Env.get().engine, future=True) as session:
            _ = [dir_record for dir_record in session.query(schema.Dir).all()]
            self.schema_objs = {
                dir_record.id: Dir(dir_record.id, dir_record.parent_id, schema.DirMd(**dir_record.md).name)
                for dir_record in session.query(schema.Dir).all()
            }

        # identify root dir
        root_dirs = [dir for dir in self.schema_objs.values() if dir._dir_id is None]
        assert len(root_dirs) == 1
        self.root_dir = root_dirs[0]

        # build dir_contents
        def record_dir(dir: Dir) -> None:
            if dir._id in self.dir_contents:
                return
            else:
                self.dir_contents[dir._id] = {}
            if dir._dir_id is not None:
                record_dir(self.schema_objs[dir._dir_id])
                self.dir_contents[dir._dir_id][dir._name] = dir

        for dir in self.schema_objs.values():
            record_dir(dir)

    def _resolve_path(self, path: Path) -> SchemaObject:
        if path.is_root:
            return self.root_dir
        dir = self.root_dir
        for i, component in enumerate(path.components):
            if component not in self.dir_contents[dir._id]:
                raise excs.Error(f'No such path: {".".join(path.components[:i + 1])}')
            schema_obj = self.dir_contents[dir._id][component]
            if i < len(path.components) - 1:
                if not isinstance(schema_obj, Dir):
                    raise excs.Error(f'Not a directory: {".".join(path.components[:i + 1])}')
                dir = schema_obj
        return schema_obj

    def __getitem__(self, path: Path) -> SchemaObject:
        return self._resolve_path(path)

    def get_schema_obj(self, id: UUID) -> Optional[SchemaObject]:
        return self.schema_objs.get(id)

    def add_schema_obj(self, dir_id: UUID, name: str, val: SchemaObject) -> None:
        self.dir_contents[dir_id][name] = val
        self.schema_objs[val._id] = val

    def __setitem__(self, path: Path, val: SchemaObject) -> None:
        parent_dir = self._resolve_path(path.parent)
        assert path.name not in self.dir_contents[parent_dir._id]
        self.schema_objs[val._id] = val
        self.dir_contents[parent_dir._id][path.name] = val
        if isinstance(val, Dir):
            self.dir_contents[val._id] = {}

    def __delitem__(self, path: Path) -> None:
        parent_dir = self._resolve_path(path.parent)
        assert path.name in self.dir_contents[parent_dir._id]
        obj = self.dir_contents[parent_dir._id][path.name]
        del self.dir_contents[parent_dir._id][path.name]
        if isinstance(obj, Dir):
            del self.dir_contents[obj._id]
        del self.schema_objs[obj._id]

    def move(self, from_path: Path, to_path: Path) -> None:
        from_dir = self._resolve_path(from_path.parent)
        assert isinstance(from_dir, Dir)
        assert from_path.name in self.dir_contents[from_dir._id]
        obj = self.dir_contents[from_dir._id][from_path.name]
        del self.dir_contents[from_dir._id][from_path.name]
        to_dir = self._resolve_path(to_path.parent)
        assert to_path.name not in self.dir_contents[to_dir._id]
        self.dir_contents[to_dir._id][to_path.name] = obj

    def check_is_valid(self, path: Path, expected: Optional[Type[SchemaObject]]) -> None:
        """Check that path is valid and that the object at path has the expected type.

        Args:
            path: path to check
            expected: expected type of object at path or None if object should not exist

        Raises:
            Error if path is invalid or object at path has wrong type
        """
        # check for existence
        if expected is not None:
            schema_obj = self._resolve_path(path)
            if not isinstance(schema_obj, expected):
                raise excs.Error(
                    f'{str(path)} needs to be a {expected.display_name()} but is a {type(schema_obj).display_name()}')
        if expected is None:
            parent_obj = self._resolve_path(path.parent)
            if not isinstance(parent_obj, Dir):
                raise excs.Error(
                    f'{str(path.parent)} is a {type(parent_obj).display_name()}, not a {Dir.display_name()}')
            if path.name in self.dir_contents[parent_obj._id]:
                obj = self.dir_contents[parent_obj._id][path.name]
                raise excs.Error(f"{type(obj).display_name()} '{str(path)}' already exists")

    def get_children(self, parent: Path, child_type: Optional[Type[SchemaObject]], recursive: bool) -> List[Path]:
        dir = self._resolve_path(parent)
        if not isinstance(dir, Dir):
            raise excs.Error(f'{str(parent)} is a {type(dir).display_name()}, not a directory')
        matches = [
            obj for obj in self.dir_contents[dir._id].values() if child_type is None or isinstance(obj, child_type)
        ]
        result = [copy.copy(parent).append(obj._name) for obj in matches]
        if recursive:
            for dir in [obj for obj in self.dir_contents[dir._id].values() if isinstance(obj, Dir)]:
                result.extend(self.get_children(copy.copy(parent).append(dir._name), child_type, recursive))
        return result




from __future__ import annotations

import logging
import sys
from typing import List, Optional, Any

import pixeltable.catalog as catalog
import pixeltable.exceptions as excs
import pixeltable.exprs as exprs
from .data_row_batch import DataRowBatch
from .exec_node import ExecNode

_logger = logging.getLogger('pixeltable')

class AggregationNode(ExecNode):
    def __init__(
            self, tbl: catalog.TableVersion, row_builder: exprs.RowBuilder, group_by: List[exprs.Expr],
            agg_fn_calls: List[exprs.FunctionCall], input_exprs: List[exprs.Expr], input: ExecNode
    ):
        super().__init__(row_builder, group_by + agg_fn_calls, input_exprs, input)
        self.input = input
        self.group_by = group_by
        self.input_exprs = input_exprs
        self.agg_fn_calls = agg_fn_calls
        self.agg_fn_eval_ctx = row_builder.create_eval_ctx(agg_fn_calls, exclude=input_exprs)
        self.output_batch = DataRowBatch(tbl, row_builder, 0)

    def _reset_agg_state(self, row_num: int) -> None:
        for fn_call in self.agg_fn_calls:
            try:
                fn_call.reset_agg()
            except Exception as e:
                _, _, exc_tb = sys.exc_info()
                expr_msg = f'init() function of the aggregate {fn_call}'
                raise excs.ExprEvalError(fn_call, expr_msg, e, exc_tb, [], row_num)

    def _update_agg_state(self, row: exprs.DataRow, row_num: int) -> None:
        for fn_call in self.agg_fn_calls:
            try:
                fn_call.update(row)
            except Exception as e:
                _, _, exc_tb = sys.exc_info()
                expr_msg = f'update() function of the aggregate {fn_call}'
                input_vals = [row[d.slot_idx] for d in fn_call.dependencies()]
                raise excs.ExprEvalError(fn_call, expr_msg, e, exc_tb, input_vals, row_num)

    def __next__(self) -> DataRowBatch:
        if self.output_batch is None:
            raise StopIteration

        prev_row: Optional[exprs.DataRow] = None
        current_group: Optional[List[Any]] = None  # the values of the group-by exprs
        num_input_rows = 0
        for row_batch in self.input:
            num_input_rows += len(row_batch)
            for row in row_batch:
                group = [row[e.slot_idx] for e in self.group_by]
                if current_group is None:
                    current_group = group
                    self._reset_agg_state(0)
                if group != current_group:
                    # we're entering a new group, emit a row for the previous one
                    self.row_builder.eval(prev_row, self.agg_fn_eval_ctx, profile=self.ctx.profile)
                    self.output_batch.add_row(prev_row)
                    current_group = group
                    self._reset_agg_state(0)
                self._update_agg_state(row, 0)
                prev_row = row
        # emit the last group
        self.row_builder.eval(prev_row, self.agg_fn_eval_ctx, profile=self.ctx.profile)
        self.output_batch.add_row(prev_row)

        result = self.output_batch
        result.flush_imgs(None, self.stored_img_cols, self.flushed_img_slots)
        self.output_batch = None
        _logger.debug(f'AggregateNode: consumed {num_input_rows} rows, returning {len(result.rows)} rows')
        return result




import datetime
import glob
import json
import os
from pathlib import Path
from typing import Dict, Any, List, Optional

import numpy as np
import pandas as pd
import pytest

import pixeltable as pxt
import pixeltable.type_system as ts
from pixeltable import catalog
from pixeltable.dataframe import DataFrameResultSet
from pixeltable.env import Env
from pixeltable.type_system import \
    ColumnType, StringType, IntType, FloatType, ArrayType, BoolType, TimestampType, JsonType, ImageType, VideoType


def make_default_type(t: ColumnType.Type) -> ColumnType:
    if t == ColumnType.Type.STRING:
        return StringType()
    if t == ColumnType.Type.INT:
        return IntType()
    if t == ColumnType.Type.FLOAT:
        return FloatType()
    if t == ColumnType.Type.BOOL:
        return BoolType()
    if t == ColumnType.Type.TIMESTAMP:
        return TimestampType()
    assert False

def make_tbl(cl: pxt.Client, name: str = 'test', col_names: Optional[List[str]] = None) -> catalog.InsertableTable:
    if col_names is None:
        col_names = ['c1']
    schema: Dict[str, ts.ColumnType] = {}
    for i, col_name in enumerate(col_names):
        schema[f'{col_name}'] = make_default_type(ColumnType.Type(i % 5))
    return cl.create_table(name, schema)

def create_table_data(t: catalog.Table, col_names: Optional[List[str]] = None, num_rows: int = 10) -> List[Dict[str, Any]]:
    if col_names is None:
        col_names = []
    data: Dict[str, Any] = {}

    sample_dict = {
        'detections': [{
            'id': '637e8e073b28441a453564cf',
            'attributes': {},
            'tags': [],
            'label': 'potted plant',
            'bounding_box': [
                0.37028125,
                0.3345305164319249,
                0.038593749999999996,
                0.16314553990610328,
            ],
            'mask': None,
            'confidence': None,
            'index': None,
            'supercategory': 'furniture',
            'iscrowd': 0,
        }, {
            'id': '637e8e073b28441a453564cf',
            'attributes': {},
            'tags': [],
            'label': 'potted plant',
            'bounding_box': [
                0.37028125,
                0.3345305164319249,
                0.038593749999999996,
                0.16314553990610328,
            ],
            'mask': None,
            'confidence': None,
            'index': None,
            'supercategory': 'furniture',
            'iscrowd': 0,
        }]
    }

    if len(col_names) == 0:
        col_names = [c.name for c in t.columns() if not c.is_computed]

    col_types = t.column_types()
    for col_name in col_names:
        col_type = col_types[col_name]
        col_data: Any = None
        if col_type.is_string_type():
            col_data = ['test string'] * num_rows
        if col_type.is_int_type():
            col_data = np.random.randint(0, 100, size=num_rows).tolist()
        if col_type.is_float_type():
            col_data = (np.random.random(size=num_rows) * 100).tolist()
        if col_type.is_bool_type():
            col_data = np.random.randint(0, 2, size=num_rows)
            col_data = [False if i == 0 else True for i in col_data]
        if col_type.is_timestamp_type():
            col_data = [datetime.datetime.now()] * num_rows
        if col_type.is_json_type():
            col_data = [sample_dict] * num_rows
        if col_type.is_array_type():
            col_data = [np.ones(col_type.shape, dtype=col_type.numpy_dtype()) for i in range(num_rows)]
        if col_type.is_image_type():
            image_path = get_image_files()[0]
            col_data = [image_path for i in range(num_rows)]
        if col_type.is_video_type():
            video_path = get_video_files()[0]
            col_data = [video_path for i in range(num_rows)]
        data[col_name] = col_data
    rows = [{col_name: data[col_name][i] for col_name in col_names} for i in range(num_rows)]
    return rows

def create_test_tbl(client: pxt.Client, name: str = 'test_tbl') -> catalog.Table:
    schema = {
        'c1': StringType(nullable=False),
        'c1n': StringType(nullable=True),
        'c2': IntType(nullable=False),
        'c3': FloatType(nullable=False),
        'c4': BoolType(nullable=False),
        'c5': TimestampType(nullable=False),
        'c6': JsonType(nullable=False),
        'c7': JsonType(nullable=False),
    }
    t = client.create_table(name, schema, primary_key='c2')
    t.add_column(c8=[[1, 2, 3], [4, 5, 6]])

    num_rows = 100
    d1 = {
        'f1': 'test string 1',
        'f2': 1,
        'f3': 1.0,
        'f4': True,
        'f5': [1.0, 2.0, 3.0, 4.0],
        'f6': {
            'f7': 'test string 2',
            'f8': [1.0, 2.0, 3.0, 4.0],
        },
    }
    d2 = [d1, d1]

    c1_data = [f'test string {i}' for i in range(num_rows)]
    c2_data = [i for i in range(num_rows)]
    c3_data = [float(i) for i in range(num_rows)]
    c4_data = [bool(i % 2) for i in range(num_rows)]
    c5_data = [datetime.datetime.now()] * num_rows
    c6_data = []
    for i in range(num_rows):
        d = {
            'f1': f'test string {i}',
            'f2': i,
            'f3': float(i),
            'f4': bool(i % 2),
            'f5': [1.0, 2.0, 3.0, 4.0],
            'f6': {
                'f7': 'test string 2',
                'f8': [1.0, 2.0, 3.0, 4.0],
            },
        }
        c6_data.append(d)

    c7_data = [d2] * num_rows
    rows = [
        {
            'c1': c1_data[i],
            'c1n': c1_data[i] if i % 10 != 0 else None,
            'c2': c2_data[i],
            'c3': c3_data[i],
            'c4': c4_data[i],
            'c5': c5_data[i],
            'c6': c6_data[i],
            'c7': c7_data[i],
        }
        for i in range(num_rows)
    ]
    t.insert(rows)
    return t

def create_all_datatypes_tbl(test_client: pxt.Client) -> catalog.Table:
    """ Creates a table with all supported datatypes.
    """
    schema = {
        'row_id': IntType(nullable=False), # used for row selection
        'c_array': ArrayType(shape=(10,),  dtype=FloatType(), nullable=True),
        'c_bool': BoolType(nullable=True),
        'c_float': FloatType(nullable=True),
        'c_image': ImageType(nullable=True),
        'c_int': IntType(nullable=True),
        'c_json': JsonType(nullable=True),
        'c_string': StringType(nullable=True),
        'c_timestamp': TimestampType(nullable=True),
        'c_video': VideoType(nullable=True),
    }
    tbl = test_client.create_table('all_datatype_tbl', schema)
    example_rows = create_table_data(tbl, num_rows=11)

    for i,r in enumerate(example_rows):
        r['row_id'] = i # row_id

    tbl.insert(example_rows)
    return tbl

def read_data_file(dir_name: str, file_name: str, path_col_names: Optional[List[str]] = None) -> List[Dict[str, Any]]:
    """
    Locate dir_name, create df out of file_name.
    path_col_names: col names in csv file that contain file names; those will be converted to absolute paths
    by adding the path to 'file_name' as a prefix.
    Returns:
        tuple of (list of rows, list of column names)
    """
    if path_col_names is None:
        path_col_names = []
    tests_dir = os.path.dirname(__file__) # search with respect to tests/ dir
    glob_result = glob.glob(f'{tests_dir}/**/{dir_name}', recursive=True)
    assert len(glob_result) == 1, f'Could not find {dir_name}'
    abs_path = Path(glob_result[0])
    data_file_path = abs_path / file_name
    assert data_file_path.is_file(), f'Not a file: {str(data_file_path)}'
    df = pd.read_csv(str(data_file_path))
    for col_name in path_col_names:
        assert col_name in df.columns
        df[col_name] = df.apply(lambda r: str(abs_path / r[col_name]), axis=1)
    return df.to_dict(orient='records')

def get_video_files(include_bad_video=False) -> List[str]:
    tests_dir = os.path.dirname(__file__) # search with respect to tests/ dir
    glob_result = glob.glob(f'{tests_dir}/**/videos/*', recursive=True)
    if not include_bad_video:
        glob_result = [f for f in glob_result if 'bad_video' not in f]
    return glob_result

def get_image_files() -> List[str]:
    tests_dir = os.path.dirname(__file__) # search with respect to tests/ dir
    glob_result = glob.glob(f'{tests_dir}/**/imagenette2-160/*', recursive=True)
    return glob_result

def get_audio_files(include_bad_audio=False) -> List[str]:
    tests_dir = os.path.dirname(__file__)
    glob_result = glob.glob(f'{tests_dir}/**/audio/*', recursive=True)
    if not include_bad_audio:
        glob_result = [f for f in glob_result if 'bad_audio' not in f]
    return glob_result

def get_documents() -> List[str]:
    tests_dir = os.path.dirname(__file__)
    # for now, we can only handle .html and .md
    return [p for p in glob.glob(f'{tests_dir}/**/documents/*', recursive=True) if not p.endswith('.pdf')]

def get_sentences(n: int = 100) -> List[str]:
    tests_dir = os.path.dirname(__file__)
    path = glob.glob(f'{tests_dir}/**/jeopardy.json', recursive=True)[0]
    with open(path, 'r') as f:
        questions_list = json.load(f)
    # this dataset contains \' around the questions
    return [q['question'].replace("'", '') for q in questions_list[:n]]

def assert_resultset_eq(r1: DataFrameResultSet, r2: DataFrameResultSet) -> None:
    assert len(r1) == len(r2)
    assert len(r1.column_names()) == len(r2.column_names())  # we don't care about the actual column names
    r1_pd = r1.to_pandas()
    r2_pd = r2.to_pandas()
    for i in range(len(r1.column_names())):
        # only compare column values
        s1 = r1_pd.iloc[:, i]
        s2 = r2_pd.iloc[:, i]
        if s1.dtype == np.float64:
            assert np.allclose(s1, s2)
        else:
            assert s1.equals(s2)

def skip_test_if_not_installed(package) -> None:
    if not Env.get().is_installed_package(package):
        pytest.skip(f'Package `{package}` is not installed.')



import json
import logging
import os
import pathlib
from typing import List

import numpy as np
import pytest

import pixeltable as pxt
import pixeltable.catalog as catalog
from pixeltable import exprs
from pixeltable import functions as ptf
from pixeltable.exprs import RELATIVE_PATH_ROOT as R
from pixeltable.metadata import SystemInfo, create_system_info
from pixeltable.metadata.schema import TableSchemaVersion, TableVersion, Table, Function, Dir
from pixeltable.tests.utils import read_data_file, create_test_tbl, create_all_datatypes_tbl, skip_test_if_not_installed
from pixeltable.type_system import StringType, ImageType, FloatType


@pytest.fixture(scope='session')
def init_env(tmp_path_factory) -> None:
    from pixeltable.env import Env
    # set the relevant env vars for Client() to connect to the test db

    shared_home = pathlib.Path(os.environ.get('PIXELTABLE_HOME', '~/.pixeltable')).expanduser()
    home_dir = str(tmp_path_factory.mktemp('base') / '.pixeltable')
    os.environ['PIXELTABLE_HOME'] = home_dir
    os.environ['PIXELTABLE_CONFIG'] = str(shared_home / 'config.yaml')
    test_db = 'test'
    os.environ['PIXELTABLE_DB'] = test_db
    os.environ['PIXELTABLE_PGDATA'] = str(shared_home / 'pgdata')

    # this also runs create_all()
    Env.get().set_up(echo=True)
    yield
    # leave db in place for debugging purposes

@pytest.fixture(scope='function')
def test_client(init_env) -> pxt.Client:
    # Clean the DB *before* instantiating a client object. This is because some tests
    # (such as test_migration.py) may leave the DB in a broken state, from which the
    # client is uninstantiable.
    clean_db()
    cl = pxt.Client(reload=True)
    cl.logging(level=logging.DEBUG, to_stdout=True)
    yield cl


def clean_db(restore_tables: bool = True) -> None:
    from pixeltable.env import Env
    # The logic from Client.reset_catalog() has been moved here, so that it
    # does not rely on instantiating a Client object. As before, UUID-named data tables will
    # not be cleaned. If in the future it is desirable to clean out data tables as well,
    # the commented lines may be used to drop ALL tables from the test db.
    # sql_md = declarative_base().metadata
    # sql_md.reflect(Env.get().engine)
    # sql_md.drop_all(bind=Env.get().engine)
    engine = Env.get().engine
    SystemInfo.__table__.drop(engine, checkfirst=True)
    TableSchemaVersion.__table__.drop(engine, checkfirst=True)
    TableVersion.__table__.drop(engine, checkfirst=True)
    Table.__table__.drop(engine, checkfirst=True)
    Function.__table__.drop(engine, checkfirst=True)
    Dir.__table__.drop(engine, checkfirst=True)
    if restore_tables:
        Dir.__table__.create(engine)
        Function.__table__.create(engine)
        Table.__table__.create(engine)
        TableVersion.__table__.create(engine)
        TableSchemaVersion.__table__.create(engine)
        SystemInfo.__table__.create(engine)
        create_system_info(engine)


@pytest.fixture(scope='function')
def test_tbl(test_client: pxt.Client) -> catalog.Table:
    return create_test_tbl(test_client)

# @pytest.fixture(scope='function')
# def test_stored_fn(test_client: pt.Client) -> pt.Function:
#     @pt.udf(return_type=pt.IntType(), param_types=[pt.IntType()])
#     def test_fn(x):
#         return x + 1
#     test_client.create_function('test_fn', test_fn)
#     return test_fn

@pytest.fixture(scope='function')
def test_tbl_exprs(test_tbl: catalog.Table) -> List[exprs.Expr]:
#def test_tbl_exprs(test_tbl: catalog.Table, test_stored_fn: pt.Function) -> List[exprs.Expr]:

    t = test_tbl
    return [
        t.c1,
        t.c7['*'].f1,
        exprs.Literal('test'),
        exprs.InlineDict({
            'a': t.c1, 'b': t.c6.f1, 'c': 17,
            'd': exprs.InlineDict({'e': t.c2}),
            'f': exprs.InlineArray((t.c3, t.c3))
        }),
        exprs.InlineArray([[t.c2, t.c2], [t.c2, t.c2]]),
        t.c2 > 5,
        t.c2 == None,
        ~(t.c2 > 5),
        (t.c2 > 5) & (t.c1 == 'test'),
        (t.c2 > 5) | (t.c1 == 'test'),
        t.c7['*'].f5 >> [R[3], R[2], R[1], R[0]],
        t.c8[0, 1:],
        t.c2.astype(FloatType()),
        (t.c2 + 1).astype(FloatType()),
        t.c2.apply(str),
        (t.c2 + 1).apply(str),
        t.c3.apply(str),
        t.c4.apply(str),
        t.c5.apply(str),
        t.c6.apply(str),
        t.c1.apply(json.loads),
        t.c8.errortype,
        t.c8.errormsg,
        ptf.sum(t.c2, group_by=t.c4, order_by=t.c3),
        #test_stored_fn(t.c2),
    ]

@pytest.fixture(scope='function')
def all_datatypes_tbl(test_client: pxt.Client) -> catalog.Table:
    return create_all_datatypes_tbl(test_client)

@pytest.fixture(scope='function')
def img_tbl(test_client: pxt.Client) -> catalog.Table:
    schema = {
        'img': ImageType(nullable=False),
        'category': StringType(nullable=False),
        'split': StringType(nullable=False),
    }
    # this table is not indexed in order to avoid the cost of computing embeddings
    tbl = test_client.create_table('test_img_tbl', schema)
    rows = read_data_file('imagenette2-160', 'manifest.csv', ['img'])
    tbl.insert(rows)
    return tbl

@pytest.fixture(scope='function')
def img_tbl_exprs(img_tbl: catalog.Table) -> List[exprs.Expr]:
    img_t = img_tbl
    return [
        img_t.img.width,
        img_t.img.rotate(90),
        # we're using a list here, not a tuple; the latter turns into a list during the back/forth conversion
        img_t.img.rotate(90).resize([224, 224]),
        img_t.img.fileurl,
        img_t.img.localpath,
    ]

# TODO: why does this not work with a session scope? (some user tables don't get created with create_all())
#@pytest.fixture(scope='session')
#def indexed_img_tbl(init_env: None) -> catalog.Table:
#    cl = pt.Client()
#    db = cl.create_db('test_indexed')
@pytest.fixture(scope='function')
def indexed_img_tbl(test_client: pxt.Client) -> catalog.Table:
    skip_test_if_not_installed('nos')
    cl = test_client
    schema = {
        'img': { 'type': ImageType(nullable=False), 'indexed': True },
        'category': StringType(nullable=False),
        'split': StringType(nullable=False),
    }
    tbl = cl.create_table('test_indexed_img_tbl', schema)
    rows = read_data_file('imagenette2-160', 'manifest.csv', ['img'])
    # select output_rows randomly in the hope of getting a good sample of the available categories
    rng = np.random.default_rng(17)
    idxs = rng.choice(np.arange(len(rows)), size=40, replace=False)
    rows = [rows[i] for i in idxs]
    tbl.insert(rows)
    return tbl



import base64

import PIL.Image

from pixeltable.type_system import ImageType, StringType
import pixeltable.func as func


@func.udf
def b64_encode(img: PIL.Image.Image, image_format: str = 'png') -> str:
    # Encode this image as a b64-encoded png.
    import io
    bytes_arr = io.BytesIO()
    img.save(bytes_arr, format=image_format)
    b64_bytes = base64.b64encode(bytes_arr.getvalue())
    return b64_bytes.decode('utf-8')



from __future__ import annotations
from typing import Optional, List, Any, Dict, Tuple
import copy

import sqlalchemy as sql

from .expr import Expr, ExprScope
from .json_mapper import JsonMapper
from .data_row import DataRow
from .row_builder import RowBuilder
import pixeltable.type_system as ts


class ObjectRef(Expr):
    """
    Reference to an intermediate result, such as the "scope variable" produced by a JsonMapper.
    The object is generated/materialized elsewhere and establishes a new scope.
    """
    def __init__(self, scope: ExprScope, owner: JsonMapper):
        # TODO: do we need an Unknown type after all?
        super().__init__(ts.JsonType())  # JsonType: this could be anything
        self._scope = scope
        self.owner = owner
        self.id = self._create_id()

    def scope(self) -> ExprScope:
        return self._scope

    def __str__(self) -> str:
        assert False

    def _equals(self, other: ObjectRef) -> bool:
        return self.owner is other.owner

    def sql_expr(self) -> Optional[sql.ClauseElement]:
        return None

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        # this will be called, but the value has already been materialized elsewhere
        pass




from __future__ import annotations

import dataclasses
import json
import logging
from pathlib import Path
from typing import Union, Any, List, Dict, Optional, Callable, Set, Tuple
from uuid import UUID

import pandas as pd
import sqlalchemy as sql

import pixeltable
import pixeltable.catalog as catalog
import pixeltable.env as env
import pixeltable.exceptions as excs
import pixeltable.exprs as exprs
import pixeltable.metadata.schema as schema
import pixeltable.type_system as ts
from .column import Column
from .globals import is_valid_identifier, is_system_column_name
from .schema_object import SchemaObject
from .table_version import TableVersion
from .table_version_path import TableVersionPath

_logger = logging.getLogger('pixeltable')

class Table(SchemaObject):
    """Base class for all tabular SchemaObjects."""

    @dataclasses.dataclass
    class UpdateStatus:
        num_rows: int = 0
        # TODO: change to num_computed_columns (the number of computed slots isn't really meaningful to the user)
        num_computed_values: int = 0
        num_excs: int = 0
        updated_cols: List[str] = dataclasses.field(default_factory=list)
        cols_with_excs: List[str] = dataclasses.field(default_factory=list)

    def __init__(self, id: UUID, dir_id: UUID, name: str, tbl_version_path: TableVersionPath):
        super().__init__(id, name, dir_id)
        self.is_dropped = False
        self.tbl_version_path = tbl_version_path

    def move(self, new_name: str, new_dir_id: UUID) -> None:
        super().move(new_name, new_dir_id)
        with env.Env.get().engine.begin() as conn:
            stmt = sql.text((
                f"UPDATE {schema.Table.__table__} "
                f"SET {schema.Table.dir_id.name} = :new_dir_id, "
                f"    {schema.Table.md.name}['name'] = :new_name "
                f"WHERE {schema.Table.id.name} = :id"))
            conn.execute(stmt, {'new_dir_id': new_dir_id, 'new_name': json.dumps(new_name), 'id': self._id})

    def version(self) -> int:
        """Return the version of this table. Used by tests to ascertain version changes."""
        return self.tbl_version_path.tbl_version.version

    def _tbl_version(self) -> TableVersion:
        """Return TableVersion for just this table."""
        return self.tbl_version_path.tbl_version

    def __hash__(self) -> int:
        return hash(self._tbl_version().id)

    def _check_is_dropped(self) -> None:
        if self.is_dropped:
            raise excs.Error(f'{self.display_name()} {self.name} has been dropped')

    def __getattr__(self, col_name: str) -> 'pixeltable.exprs.ColumnRef':
        """Return a ColumnRef for the given column name.
        """
        return getattr(self.tbl_version_path, col_name)

    def __getitem__(self, index: object) -> Union['pixeltable.exprs.ColumnRef', 'pixeltable.dataframe.DataFrame']:
        """Return a ColumnRef for the given column name, or a DataFrame for the given slice.
        """
        return self.tbl_version_path.__getitem__(index)

    def df(self) -> 'pixeltable.dataframe.DataFrame':
        """Return a DataFrame for this table.
        """
        # local import: avoid circular imports
        from pixeltable.dataframe import DataFrame
        return DataFrame(self.tbl_version_path)

    def select(self, *items: Any, **named_items: Any) -> 'pixeltable.dataframe.DataFrame':
        """Return a DataFrame for this table.
        """
        # local import: avoid circular imports
        from pixeltable.dataframe import DataFrame
        return DataFrame(self.tbl_version_path).select(*items, **named_items)

    def where(self, pred: 'exprs.Predicate') -> 'pixeltable.dataframe.DataFrame':
        """Return a DataFrame for this table.
        """
        # local import: avoid circular imports
        from pixeltable.dataframe import DataFrame
        return DataFrame(self.tbl_version_path).where(pred)

    def order_by(self, *items: 'exprs.Expr', asc: bool = True) -> 'pixeltable.dataframe.DataFrame':
        """Return a DataFrame for this table.
        """
        # local import: avoid circular imports
        from pixeltable.dataframe import DataFrame
        return DataFrame(self.tbl_version_path).order_by(*items, asc=asc)

    def collect(self) -> 'pixeltable.dataframe.DataFrameResultSet':  # type: ignore[name-defined, no-untyped-def]
        """Return rows from this table.
        """
        return self.df().collect()

    def show(
            self, *args, **kwargs
    ) -> 'pixeltable.dataframe.DataFrameResultSet':  # type: ignore[name-defined, no-untyped-def]
        """Return rows from this table.
        """
        return self.df().show(*args, **kwargs)

    def head(
            self, *args, **kwargs
    ) -> 'pixeltable.dataframe.DataFrameResultSet':  # type: ignore[name-defined, no-untyped-def]
        """Return the first n rows inserted into this table."""
        return self.df().head(*args, **kwargs)

    def tail(
            self, *args, **kwargs
    ) -> 'pixeltable.dataframe.DataFrameResultSet':  # type: ignore[name-defined, no-untyped-def]
        """Return the last n rows inserted into this table."""
        return self.df().tail(*args, **kwargs)

    def count(self) -> int:
        """Return the number of rows in this table."""
        return self.df().count()

    def column_names(self) -> List[str]:
        """Return the names of the columns in this table."""
        return [c.name for c in self.tbl_version_path.columns()]

    def column_types(self) -> Dict[str, ts.ColumnType]:
        """Return the names of the columns in this table."""
        return {c.name: c.col_type for c in self.tbl_version_path.columns()}

    @property
    def comment(self) -> str:
        return self.tbl_version.comment

    @comment.setter
    def comment(self, new_comment: Optional[str]):
        self.tbl_version.set_comment(new_comment)

    @property
    def num_retained_versions(self):
        return self.tbl_version.num_retained_versions

    @num_retained_versions.setter
    def num_retained_versions(self, new_num_retained_versions: int):
        self.tbl_version.set_num_retained_versions(new_num_retained_versions)

    def _description(self) -> pd.DataFrame:
        cols = self.tbl_version_path.columns()
        df = pd.DataFrame({
            'Column Name': [c.name for c in cols],
            'Type': [str(c.col_type) for c in cols],
            'Computed With': [c.value_expr.display_str(inline=False) if c.value_expr is not None else '' for c in cols],
        })
        return df

    def _description_html(self) -> pd.DataFrame:
        pd_df = self._description()
        # white-space: pre-wrap: print \n as newline
        # th: center-align headings
        return pd_df.style.set_properties(**{'white-space': 'pre-wrap', 'text-align': 'left'}) \
            .set_table_styles([dict(selector='th', props=[('text-align', 'center')])]) \
            .hide(axis='index')

    def describe(self) -> None:
        try:
            __IPYTHON__
            from IPython.display import display
            display(self._description_html())
        except NameError:
            print(self.__repr__())

    # TODO: Display comments in _repr_html()
    def __repr__(self) -> str:
        description_str = self._description().to_string(index=False)
        if self.comment is None:
            comment = ''
        else:
            comment = f'{self.comment}\n'
        return f'{self.display_name()} \'{self._name}\'\n{comment}{description_str}'

    def _repr_html_(self) -> str:
        return self._description_html()._repr_html_()

    def _drop(self) -> None:
        self._check_is_dropped()
        self.tbl_version_path.tbl_version.drop()
        self.is_dropped = True
        # update catalog
        cat = catalog.Catalog.get()
        del cat.tbls[self._id]

    # TODO Factor this out into a separate module.
    # The return type is unresolvable, but torch can't be imported since it's an optional dependency.
    def to_pytorch_dataset(self, image_format : str = 'pt') -> 'torch.utils.data.IterableDataset':
        """Return a PyTorch Dataset for this table.
            See DataFrame.to_pytorch_dataset()
        """
        from pixeltable.dataframe import DataFrame
        return DataFrame(self.tbl_version_path).to_pytorch_dataset(image_format=image_format)

    def to_coco_dataset(self) -> Path:
        """Return the path to a COCO json file for this table.
            See DataFrame.to_coco_dataset()
        """
        from pixeltable.dataframe import DataFrame
        return DataFrame(self.tbl_version_path).to_coco_dataset()

    def __setitem__(self, column_name: str, value: Union[ts.ColumnType, exprs.Expr, Callable, dict]) -> None:
        """Adds a column to the table
        Args:
            column_name: the name of the new column
            value: column type or value expression or column specification dictionary:
                column type: a Pixeltable column type (if the table already contains rows, it must be nullable)
                value expression: a Pixeltable expression that computes the column values
                column specification: a dictionary with possible keys 'type', 'value', 'stored', 'indexed'
        Examples:
            Add an int column with ``None`` values:

            >>> tbl['new_col'] = IntType(nullable=True)

            For a table with int column ``int_col``, add a column that is the factorial of ``int_col``. The names of
            the parameters of the Callable must correspond to existing column names (the column values are then passed
            as arguments to the Callable). In this case, the return type cannot be inferred and needs to be specified
            explicitly:

            >>> tbl['factorial'] = {'value': lambda int_col: math.factorial(int_col), 'type': IntType()}

            For a table with an image column ``frame``, add an image column ``rotated`` that rotates the image by
            90 degrees. In this case, the column type is inferred from the expression. Also, the column is not stored
            (by default, computed image columns are not stored but recomputed on demand):

            >>> tbl['rotated'] = tbl.frame.rotate(90)

            Do the same, but now the column is stored:

            >>> tbl['rotated'] = {'value': tbl.frame.rotate(90), 'stored': True}

            Add a resized version of the ``frame`` column and index it. The column does not need to be stored in order
            to be indexed:

            >>> tbl['small_frame'] = {'value': tbl.frame.resize([224, 224]), 'indexed': True}
        """
        if not isinstance(column_name, str):
            raise excs.Error(f'Column name must be a string, got {type(column_name)}')
        if not is_valid_identifier(column_name):
            raise excs.Error(f'Invalid column name: {column_name!r}')

        new_col = self._create_columns({column_name: value})[0]
        self._verify_column(new_col, self.column_names())
        return self.tbl_version_path.tbl_version.add_column(new_col)

    def add_column(
            self, *,
            type: Optional[ts.ColumnType] = None, stored: Optional[bool] = None, indexed: Optional[bool] = None,
            print_stats: bool = False, **kwargs: Any
    ) -> UpdateStatus:
        """Adds a column to the table.

        Args:
            kwargs: Exactly one keyword argument of the form ``column-name=type|value-expression``.
            type: The type of the column. Only valid and required if ``value-expression`` is a Callable.
            stored: Whether the column is materialized and stored or computed on demand. Only valid for image columns.
            indexed: Whether the column is indexed.
            print_stats: If ``True``, print execution metrics.

        Returns:
            execution status

        Raises:
            Error: If the column name is invalid or already exists.

        Examples:
            Add an int column with ``None`` values:

            >>> tbl.add_column(new_col=IntType())

            Alternatively, this can also be expressed as:

            >>> tbl['new_col'] = IntType()

            For a table with int column ``int_col``, add a column that is the factorial of ``int_col``. The names of
            the parameters of the Callable must correspond to existing column names (the column values are then passed
            as arguments to the Callable). In this case, the column type needs to be specified explicitly:

            >>> tbl.add_column(factorial=lambda int_col: math.factorial(int_col), type=IntType())

            Alternatively, this can also be expressed as:

            >>> tbl['factorial'] = {'value': lambda int_col: math.factorial(int_col), 'type': IntType()}

            For a table with an image column ``frame``, add an image column ``rotated`` that rotates the image by
            90 degrees. In this case, the column type is inferred from the expression. Also, the column is not stored
            (by default, computed image columns are not stored but recomputed on demand):

            >>> tbl.add_column(rotated=tbl.frame.rotate(90))

            Alternatively, this can also be expressed as:

            >>> tbl['rotated'] = tbl.frame.rotate(90)

            Do the same, but now the column is stored:

            >>> tbl.add_column(rotated=tbl.frame.rotate(90), stored=True)

            Alternatively, this can also be expressed as:

            >>> tbl['rotated'] = {'value': tbl.frame.rotate(90), 'stored': True}

            Add a resized version of the ``frame`` column and index it. The column does not need to be stored in order
            to be indexed:

            >>> tbl.add_column(small_frame=tbl.frame.resize([224, 224]), indexed=True)

            Alternatively, this can also be expressed as:

            >>> tbl['small_frame'] = {'value': tbl.frame.resize([224, 224]), 'indexed': True}
        """
        self._check_is_dropped()
        # verify kwargs and construct column schema dict
        if len(kwargs) != 1:
            raise excs.Error((
                f'add_column() requires exactly one keyword argument of the form "column-name=type|value-expression", '
                f'got {len(kwargs)} instead ({", ".join(list(kwargs.keys()))})'
            ))
        col_name, spec = next(iter(kwargs.items()))
        col_schema: Dict[str, Any] = {}
        if isinstance(spec, ts.ColumnType):
            if type is not None:
                raise excs.Error(f'add_column(): keyword argument "type" is redundant')
            col_schema['type'] = spec
        else:
            if isinstance(spec, exprs.Expr) and type is not None:
                raise excs.Error(f'add_column(): keyword argument "type" is redundant')
            col_schema['value'] = spec
        if type is not None:
            col_schema['type'] = type
        if stored is not None:
            col_schema['stored'] = stored
        if indexed is not None:
            col_schema['indexed'] = indexed

        new_col = self._create_columns({col_name: col_schema})[0]
        self._verify_column(new_col, self.column_names())
        return self.tbl_version_path.tbl_version.add_column(new_col, print_stats=print_stats)

    @classmethod
    def _validate_column_spec(cls, name: str, spec: Dict[str, Any]) -> None:
        """Check integrity of user-supplied Column spec

        We unfortunately can't use something like jsonschema for validation, because this isn't strictly a JSON schema
        (on account of containing Python Callables or Exprs).
        """
        assert isinstance(spec, dict)
        valid_keys = {'type', 'value', 'stored', 'indexed'}
        has_type = False
        for k in spec.keys():
            if k not in valid_keys:
                raise excs.Error(f'Column {name}: invalid key {k!r}')

        if 'type' in spec:
            has_type = True
            if not isinstance(spec['type'], ts.ColumnType):
                raise excs.Error(f'Column {name}: "type" must be a ColumnType, got {spec["type"]}')

        if 'value' in spec:
            value_spec = spec['value']
            value_expr = exprs.Expr.from_object(value_spec)
            if value_expr is None:
                # needs to be a Callable
                if not isinstance(value_spec, Callable):
                    raise excs.Error(
                        f'Column {name}: value needs to be either a Pixeltable expression or a Callable, '
                        f'but it is a {type(value_spec)}')
                if 'type' not in spec:
                    raise excs.Error(f'Column {name}: "type" is required if value is a Callable')
            else:
                has_type = True
                if 'type' in spec:
                    raise excs.Error(f'Column {name}: "type" is redundant if value is a Pixeltable expression')

        if 'stored' in spec and not isinstance(spec['stored'], bool):
            raise excs.Error(f'Column {name}: "stored" must be a bool, got {spec["stored"]}')
        if 'indexed' in spec and not isinstance(spec['indexed'], bool):
            raise excs.Error(f'Column {name}: "indexed" must be a bool, got {spec["indexed"]}')
        if not has_type:
            raise excs.Error(f'Column {name}: "type" is required')

    @classmethod
    def _create_columns(cls, schema: Dict[str, Any]) -> List[Column]:
        """Construct list of Columns, given schema"""
        columns: List[Column] = []
        for name, spec in schema.items():
            col_type: Optional[ts.ColumnType] = None
            value_expr: Optional[exprs.Expr] = None
            stored: Optional[bool] = None
            indexed: Optional[bool] = None
            primary_key: Optional[bool] = None

            if isinstance(spec, ts.ColumnType):
                # TODO: create copy
                col_type = spec
            elif isinstance(spec, exprs.Expr):
                # create copy so we can modify it
                value_expr = spec.copy()
            elif isinstance(spec, Callable):
                raise excs.Error((
                    f'Column {name} computed with a Callable: specify using a dictionary with '
                    f'the "value" and "type" keys (e.g., "{name}": {{"value": <Callable>, "type": IntType()}})'
                ))
            elif isinstance(spec, dict):
                cls._validate_column_spec(name, spec)
                col_type = spec.get('type')
                value_expr = spec.get('value')
                if value_expr is not None and isinstance(value_expr, exprs.Expr):
                    # create copy so we can modify it
                    value_expr = value_expr.copy()
                stored = spec.get('stored')
                indexed = spec.get('indexed')
                primary_key = spec.get('primary_key')

            column = Column(
                name, col_type=col_type, computed_with=value_expr, stored=stored, indexed=indexed,
                primary_key=primary_key)
            columns.append(column)
        return columns

    @classmethod
    def _verify_column(cls, col: Column, existing_column_names: Set[str]) -> None:
        """Check integrity of user-supplied Column and supply defaults"""
        if is_system_column_name(col.name):
            raise excs.Error(f'Column name {col.name} is reserved')
        if not is_valid_identifier(col.name):
            raise excs.Error(f"Invalid column name: '{col.name}'")
        if col.name in existing_column_names:
            raise excs.Error(f'Duplicate column name: {col.name}')
        if col.stored is False and not (col.is_computed and col.col_type.is_image_type()):
            raise excs.Error(f'Column {col.name}: stored={col.stored} only applies to computed image columns')
        if col.stored is False and not (col.col_type.is_image_type() and not col.has_window_fn_call()):
            raise excs.Error((
                f'Column {col.name}: stored={col.stored} is not valid for image columns computed with a streaming '
                f'function'))
        if col.stored is None:
            col.stored = not (col.is_computed and col.col_type.is_image_type() and not col.has_window_fn_call())

    @classmethod
    def _verify_schema(cls, schema: List[Column]) -> None:
        """Check integrity of user-supplied schema and set defaults"""
        column_names: Set[str] = set()
        for col in schema:
            cls._verify_column(col, column_names)
            column_names.add(col.name)

    def drop_column(self, name: str) -> None:
        """Drop a column from the table.

        Args:
            name: The name of the column to drop.

        Raises:
            Error: If the column does not exist or if it is referenced by a computed column.

        Examples:
            Drop column ``factorial``:

            >>> tbl.drop_column('factorial')
        """
        self._check_is_dropped()
        self.tbl_version_path.tbl_version.drop_column(name)

    def rename_column(self, old_name: str, new_name: str) -> None:
        """Rename a column.

        Args:
            old_name: The current name of the column.
            new_name: The new name of the column.

        Raises:
            Error: If the column does not exist or if the new name is invalid or already exists.

        Examples:
            Rename column ``factorial`` to ``fac``:

            >>> tbl.rename_column('factorial', 'fac')
        """
        self._check_is_dropped()
        self.tbl_version_path.tbl_version.rename_column(old_name, new_name)

    def update(
            self, value_spec: Dict[str, Union['pixeltable.exprs.Expr', Any]],
            where: Optional['pixeltable.exprs.Predicate'] = None, cascade: bool = True
    ) -> UpdateStatus:
        """Update rows in this table.

        Args:
            value_spec: a dictionary mapping column names to literal values or Pixeltable expressions.
            where: a Predicate to filter rows to update.
            cascade: if True, also update all computed columns that transitively depend on the updated columns.

        Examples:
            Set newly-added column `int_col` to 1 for all rows:

            >>> tbl.update({'int_col': 1})

            Set newly-added column `int_col` to 1 for all rows where `int_col` is 0:

            >>> tbl.update({'int_col': 1}, where=tbl.int_col == 0)

            Set `int_col` to the value of `other_int_col` + 1:

            >>> tbl.update({'int_col': tbl.other_int_col + 1})

            Increment `int_col` by 1 for all rows where `int_col` is 0:

            >>> tbl.update({'int_col': tbl.int_col + 1}, where=tbl.int_col == 0)
        """
        from pixeltable import exprs
        update_targets: List[Tuple[Column, exprs.Expr]] = []
        for col_name, val in value_spec.items():
            if not isinstance(col_name, str):
                raise excs.Error(f'Update specification: dict key must be column name, got {col_name!r}')
            col = self.tbl_version_path.get_column(col_name, include_bases=False)
            if col is None:
                # TODO: return more informative error if this is trying to update a base column
                raise excs.Error(f'Column {col_name} unknown')
            if col.is_computed:
                raise excs.Error(f'Column {col_name} is computed and cannot be updated')
            if col.primary_key:
                raise excs.Error(f'Column {col_name} is a primary key column and cannot be updated')
            if col.col_type.is_media_type():
                raise excs.Error(f'Column {col_name} has type image/video/audio/document and cannot be updated')

            # make sure that the value is compatible with the column type
            # check if this is a literal
            try:
                value_expr = exprs.Literal(val, col_type=col.col_type)
            except TypeError:
                # it's not a literal, let's try to create an expr from it
                value_expr = exprs.Expr.from_object(val)
                if value_expr is None:
                    raise excs.Error(f'Column {col_name}: value {val!r} is not a recognized literal or expression')
                if not col.col_type.matches(value_expr.col_type):
                    raise excs.Error((
                        f'Type of value {val!r} ({value_expr.col_type}) is not compatible with the type of column '
                        f'{col_name} ({col.col_type})'
                    ))
            update_targets.append((col, value_expr))

        from pixeltable.plan import Planner
        if where is not None:
            if not isinstance(where, exprs.Predicate):
                raise excs.Error(f"'where' argument must be a Predicate, got {type(where)}")
            analysis_info = Planner.analyze(self.tbl_version_path, where)
            if analysis_info.similarity_clause is not None:
                raise excs.Error('nearest() cannot be used with update()')
            # for now we require that the updated rows can be identified via SQL, rather than via a Python filter
            if analysis_info.filter is not None:
                raise excs.Error(f'Filter {analysis_info.filter} not expressible in SQL')

        return self.tbl_version_path.tbl_version.update(update_targets, where, cascade)

    def revert(self) -> None:
        """Reverts the table to the previous version.

        .. warning::
            This operation is irreversible.
        """
        self._check_is_dropped()
        self.tbl_version_path.tbl_version.revert()



import tempfile
from pathlib import Path
from typing import Optional, Union

import PIL.Image
import av
import av.container
import av.stream
import numpy as np

import pixeltable.env as env
import pixeltable.func as func
# import all standard function modules here so they get registered with the FunctionRegistry
import pixeltable.functions.pil.image
from pixeltable import exprs
from pixeltable.type_system import IntType, ColumnType, FloatType, ImageType, VideoType
# automatically import all submodules so that the udfs get registered
from . import image, string, video, openai, together, fireworks, huggingface

# TODO: remove and replace calls with astype()
def cast(expr: exprs.Expr, target_type: ColumnType) -> exprs.Expr:
    expr.col_type = target_type
    return expr

@func.uda(
    update_types=[IntType()], value_type=IntType(), name='sum', allows_window=True, requires_order_by=False)
class SumAggregator(func.Aggregator):
    def __init__(self):
        self.sum: Union[int, float] = 0
    def update(self, val: Union[int, float]) -> None:
        if val is not None:
            self.sum += val
    def value(self) -> Union[int, float]:
        return self.sum


@func.uda(
    update_types=[IntType()], value_type=IntType(), name='count', allows_window = True, requires_order_by = False)
class CountAggregator(func.Aggregator):
    def __init__(self):
        self.count = 0
    def update(self, val: int) -> None:
        if val is not None:
            self.count += 1
    def value(self) -> int:
        return self.count


@func.uda(
    update_types=[IntType()], value_type=FloatType(), name='mean', allows_window=False, requires_order_by=False)
class MeanAggregator(func.Aggregator):
    def __init__(self):
        self.sum = 0
        self.count = 0
    def update(self, val: int) -> None:
        if val is not None:
            self.sum += val
            self.count += 1
    def value(self) -> float:
        if self.count == 0:
            return None
        return self.sum / self.count


@func.uda(
    init_types=[IntType()], update_types=[ImageType()], value_type=VideoType(), name='make_video',
    requires_order_by=True, allows_window=False)
class VideoAggregator(func.Aggregator):
    def __init__(self, fps: int = 25):
        """follows https://pyav.org/docs/develop/cookbook/numpy.html#generating-video"""
        self.container: Optional[av.container.OutputContainer] = None
        self.stream: Optional[av.stream.Stream] = None
        self.fps = fps

    def update(self, frame: PIL.Image.Image) -> None:
        if frame is None:
            return
        if self.container is None:
            (_, output_filename) = tempfile.mkstemp(suffix='.mp4', dir=str(env.Env.get().tmp_dir))
            self.out_file = Path(output_filename)
            self.container = av.open(str(self.out_file), mode='w')
            self.stream = self.container.add_stream('h264', rate=self.fps)
            self.stream.pix_fmt = 'yuv420p'
            self.stream.width = frame.width
            self.stream.height = frame.height

        av_frame = av.VideoFrame.from_ndarray(np.array(frame.convert('RGB')), format='rgb24')
        for packet in self.stream.encode(av_frame):
            self.container.mux(packet)

    def value(self) -> str:
        for packet in self.stream.encode():
            self.container.mux(packet)
        self.container.close()
        return str(self.out_file)



from __future__ import annotations

import abc
import importlib
import inspect
from typing import Optional, Any, Type, List, Dict
import itertools

import pixeltable.exceptions as excs
import pixeltable.type_system as ts
from .function import Function
from .signature import Signature, Parameter


class Aggregator(abc.ABC):
    def update(self, *args: Any, **kwargs: Any) -> None:
        pass
    def value(self) -> Any:
        pass


class AggregateFunction(Function):
    """Function interface for an aggregation operation.

    requires_order_by: if True, the first parameter to an aggregate function defines the order in which the function
    sees rows in update()
    allows_std_agg: if True, the aggregate function can be used as a standard aggregate function w/o a window
    allows_window: if True, the aggregate function can be used with a window
    """
    ORDER_BY_PARAM = 'order_by'
    GROUP_BY_PARAM = 'group_by'
    RESERVED_PARAMS = {ORDER_BY_PARAM, GROUP_BY_PARAM}

    def __init__(
            self, aggregator_class: Type[Aggregator], self_path: str,
            init_types: List[ts.ColumnType], update_types: List[ts.ColumnType], value_type: ts.ColumnType,
            requires_order_by: bool, allows_std_agg: bool, allows_window: bool):
        self.agg_cls = aggregator_class
        self.requires_order_by = requires_order_by
        self.allows_std_agg = allows_std_agg
        self.allows_window = allows_window

        # our signature is the signature of 'update', but without self,
        # plus the parameters of 'init' as keyword-only parameters
        update_params = list(inspect.signature(self.agg_cls.update).parameters.values())[1:]  # leave out self
        assert len(update_params) == len(update_types)
        init_params = [
            inspect.Parameter(p.name, inspect.Parameter.KEYWORD_ONLY, default=p.default)
            # starting at 1: leave out self
            for p in itertools.islice(inspect.signature(self.agg_cls.__init__).parameters.values(), 1, None)
        ]
        assert len(init_params) == len(init_types)
        duplicate_params = set(p.name for p in init_params) & set(p.name for p in update_params)
        if len(duplicate_params) > 0:
            raise excs.Error(
                f'__init__() and update() cannot have parameters with the same name: '
                f'{", ".join(duplicate_params)}'
            )
        py_params = update_params + init_params  # init_params are keyword-only and come last
        py_signature = inspect.Signature(py_params)

        params = [Parameter(p.name, update_types[i], p.kind, is_batched=False) for i, p in enumerate(update_params)]
        params.extend([Parameter(p.name, init_types[i], p.kind, is_batched=False) for i, p in enumerate(init_params)])
        signature = Signature(value_type, params)
        super().__init__(signature, py_signature=py_signature, self_path=self_path)
        self.init_param_names = [p.name for p in init_params]

        # make sure the signature doesn't contain reserved parameter names;
        # do this after super().__init__(), otherwise self.name is invalid
        for param in signature.parameters:
            if param.lower() in self.RESERVED_PARAMS:
                raise excs.Error(f'{self.name}(): parameter name {param} is reserved')

    def help_str(self) -> str:
        res = super().help_str()
        res += '\n\n' + inspect.getdoc(self.agg_cls.update)
        return res

    def __call__(self, *args: object, **kwargs: object) -> 'pixeltable.exprs.Expr':
        from pixeltable import exprs

        # perform semantic analysis of special parameters 'order_by' and 'group_by'
        order_by_clause: Optional[Any] = None
        if self.ORDER_BY_PARAM in kwargs:
            if self.requires_order_by:
                raise excs.Error(
                    f'{self.display_name}(): order_by invalid, this function requires the first argument to be the '
                    f'ordering expression'
                )
            if not self.allows_window:
                raise excs.Error(
                    f'{self.display_name}(): order_by invalid with an aggregate function that does not allow windows')
            order_by_clause = kwargs.pop(self.ORDER_BY_PARAM)
        elif self.requires_order_by:
            # the first argument is the order-by expr
            if len(args) == 0:
                raise excs.Error(f'{self.display_name}(): requires an ordering expression as its first argument')
            order_by_clause = args[0]
            if not isinstance(order_by_clause, exprs.Expr):
                raise excs.Error(
                    f'{self.display_name}(): the first argument needs to be a Pixeltable expression, but instead is a '
                    f'{type(order_by_clause)}'
                )
            # don't pass the first parameter on, the Function doesn't get to see it
            args = args[1:]

        group_by_clause: Optional[Any] = None
        if self.GROUP_BY_PARAM in kwargs:
            if not self.allows_window:
                raise excs.Error(
                    f'{self.display_name}(): group_by invalid with an aggregate function that does not allow windows')
            group_by_clause = kwargs.pop(self.GROUP_BY_PARAM)

        bound_args = self.py_signature.bind(*args, **kwargs)
        self.validate_call(bound_args.arguments)
        return exprs.FunctionCall(
            self, bound_args.arguments,
            order_by_clause=[order_by_clause] if order_by_clause is not None else [],
            group_by_clause=[group_by_clause] if group_by_clause is not None else [])

    def validate_call(self, bound_args: Dict[str, Any]) -> None:
        # check that init parameters are not Exprs
        # TODO: do this in the planner (check that init parameters are either constants or only refer to grouping exprs)
        import pixeltable.exprs as exprs
        for param_name in self.init_param_names:
            if param_name in bound_args and isinstance(bound_args[param_name], exprs.Expr):
                raise excs.Error(
                    f'{self.display_name}(): init() parameter {param_name} needs to be a constant, not a Pixeltable '
                    f'expression'
                )


def uda(
        *,
        value_type: ts.ColumnType,
        update_types: List[ts.ColumnType],
        init_types: Optional[List[ts.ColumnType]] = None,
        requires_order_by: bool = False, allows_std_agg: bool = True, allows_window: bool = False,
        name: Optional[str] = None
) -> Type[Aggregator]:
    """Decorator for user-defined aggregate functions.

    The decorated class must inherit from Aggregator and implement the following methods:
    - __init__(self, ...) to initialize the aggregator
    - update(self, ...) to update the aggregator with a new value
    - value(self) to return the final result

    The decorator creates an AggregateFunction instance from the class and adds it
    to the module where the class is defined.

    Parameters:
    - init_types: list of types for the __init__() parameters; must match the number of parameters
    - update_types: list of types for the update() parameters; must match the number of parameters
    - value_type: return type of the aggregator
    - requires_order_by: if True, the first parameter to the function is the order-by expression
    - allows_std_agg: if True, the function can be used as a standard aggregate function w/o a window
    - allows_window: if True, the function can be used with a window
    - name: name of the AggregateFunction instance; if None, the class name is used
    """
    if name is not None and not name.isidentifier():
        raise excs.Error(f'Invalid name: {name}')
    if init_types is None:
        init_types = []

    def decorator(cls: Type[Aggregator]) -> Type[Aggregator]:
        # validate type parameters
        num_init_params = len(inspect.signature(cls.__init__).parameters) - 1
        if num_init_params > 0:
            if len(init_types) != num_init_params:
                raise excs.Error(
                    f'init_types must be a list of {num_init_params} types, one for each parameter of __init__()')
        num_update_params = len(inspect.signature(cls.update).parameters) - 1
        if num_update_params == 0:
            raise excs.Error('update() must have at least one parameter')
        if len(update_types) != num_update_params:
            raise excs.Error(
                f'update_types must be a list of {num_update_params} types, one for each parameter of update()')
        assert value_type is not None

        # the AggregateFunction instance resides in the same module as cls
        module_path = cls.__module__
        nonlocal name
        name = name or cls.__name__
        instance_path = f'{module_path}.{name}'

        # create the corresponding AggregateFunction instance
        instance = AggregateFunction(
            cls, instance_path, init_types, update_types, value_type, requires_order_by, allows_std_agg, allows_window)
        module = importlib.import_module(module_path)
        setattr(module, name, instance)

        return cls

    return decorator



from __future__ import annotations
from typing import Iterable, Optional, List
import abc

from .data_row_batch import DataRowBatch
from .exec_context import ExecContext
import pixeltable.exprs as exprs

class ExecNode(abc.ABC):
    """Base class of all execution nodes"""
    def __init__(
            self, row_builder: exprs.RowBuilder, output_exprs: Iterable[exprs.Expr],
            input_exprs: Iterable[exprs.Expr], input: Optional[ExecNode] = None):
        self.row_builder = row_builder
        self.input = input
        # we flush all image slots that aren't part of our output but are needed to create our output
        output_slot_idxs = {e.slot_idx for e in output_exprs}
        output_dependencies = row_builder.get_dependencies(output_exprs, exclude=input_exprs)
        self.flushed_img_slots = [
            e.slot_idx for e in output_dependencies
            if e.col_type.is_image_type() and e.slot_idx not in output_slot_idxs
        ]
        self.stored_img_cols: List[exprs.ColumnSlotIdx] = []
        self.ctx: Optional[ExecContext] = None  # all nodes of a tree share the same context

    def set_ctx(self, ctx: ExecContext) -> None:
        self.ctx = ctx
        if self.input is not None:
            self.input.set_ctx(ctx)

    def set_stored_img_cols(self, stored_img_cols: List[exprs.ColumnSlotIdx]) -> None:
        self.stored_img_cols = stored_img_cols
        # propagate batch size to the source
        if self.input is not None:
            self.input.set_stored_img_cols(stored_img_cols)

    def __iter__(self):
        return self

    @abc.abstractmethod
    def __next__(self) -> DataRowBatch:
        pass

    def open(self) -> None:
        """Bottom-up initialization of nodes for execution. Must be called before __next__."""
        if self.input is not None:
            self.input.open()
        self._open()

    def close(self) -> None:
        """Frees node resources top-down after execution. Must be called after final __next__."""
        self._close()
        if self.input is not None:
            self.input.close()

    def _open(self) -> None:
        pass

    def _close(self) -> None:
        pass




from typing import Any, Dict

import numpy as np
import pytest

import pixeltable as pxt
import pixeltable.exceptions as excs
from pixeltable.tests.utils import create_test_tbl, assert_resultset_eq
from pixeltable.type_system import IntType


class TestSnapshot:
    def run_basic_test(
            self, cl: pxt.Client, tbl: pxt.Table, snap: pxt.Table, extra_items: Dict[str, Any], filter: Any,
            reload_md: bool
    ) -> None:
        tbl_path, snap_path = cl.get_path(tbl), cl.get_path(snap)
        # run the initial query against the base table here, before reloading, otherwise the filter breaks
        tbl_select_list = [tbl[col_name] for col_name in tbl.column_names()]
        tbl_select_list.extend([value_expr for _, value_expr in extra_items.items()])
        orig_resultset = tbl.select(*tbl_select_list).where(filter).order_by(tbl.c2).collect()

        if reload_md:
            # reload md
            cl = pxt.Client(reload=True)
            tbl = cl.get_table(tbl_path)
            snap = cl.get_table(snap_path)

        # view select list: base cols followed by view cols
        snap_select_list = [snap[col_name] for col_name in snap.column_names()[len(extra_items):]]
        snap_select_list.extend([snap[col_name] for col_name in extra_items.keys()])
        snap_query = snap.select(*snap_select_list).order_by(snap.c2)
        r1 = list(orig_resultset)
        r2 = list(snap_query.collect())
        assert_resultset_eq(snap_query.collect(), orig_resultset)

        # adding data to a base table doesn't change the snapshot
        rows = list(tbl.select(tbl.c1, tbl.c1n, tbl.c2, tbl.c3, tbl.c4, tbl.c5, tbl.c6, tbl.c7).collect())
        status = tbl.insert(rows)
        assert status.num_rows == len(rows)
        assert_resultset_eq(snap_query.collect(), orig_resultset)

        # update() doesn't affect the view
        status = tbl.update({'c3': tbl.c3 + 1.0})
        assert status.num_rows == tbl.count()
        assert_resultset_eq(snap_query.collect(), orig_resultset)

        # delete() doesn't affect the view
        num_tbl_rows = tbl.count()
        status = tbl.delete()
        assert status.num_rows == num_tbl_rows
        assert_resultset_eq(snap_query.collect(), orig_resultset)

        tbl.revert()  # undo delete()
        tbl.revert()  # undo update()
        tbl.revert()  # undo insert()
        # can't revert a version referenced by a snapshot
        with pytest.raises(excs.Error) as excinfo:
            tbl.revert()
        assert 'version is needed' in str(excinfo.value)

        # can't drop a table with snapshots
        with pytest.raises(excs.Error) as excinfo:
            cl.drop_table(tbl_path)
        assert snap_path in str(excinfo.value)

        cl.drop_table(snap_path)
        cl.drop_table(tbl_path)

    def test_basic(self, test_client: pxt.Client) -> None:
        cl = test_client
        cl.create_dir('main')
        cl.create_dir('snap')
        tbl_path = 'main.tbl1'
        snap_path = 'snap.snap1'

        for reload_md in [False, True]:
            for has_filter in [False, True]:
                for has_cols in [False, True]:
                    cl = pxt.Client(reload=True)
                    tbl = create_test_tbl(name=tbl_path, client=cl)
                    schema = {
                        'v1': tbl.c3 * 2.0,
                        # include a lambda to make sure that is handled correctly
                        'v2': {'value': lambda c3: c3 * 2.0, 'type': pxt.FloatType()}
                    } if has_cols else {}
                    extra_items = {'v1': tbl.c3 * 2.0, 'v2': tbl.c3 * 2.0} if has_cols else {}
                    filter = tbl.c2 < 10 if has_filter else None
                    snap = cl.create_view(snap_path, tbl, schema=schema, filter=filter, is_snapshot=True)
                    self.run_basic_test(cl, tbl, snap, extra_items=extra_items, filter=filter, reload_md=reload_md)

    def test_views_of_snapshots(self, test_client: pxt.Client) -> None:
        cl = test_client
        t = cl.create_table('tbl', {'a': IntType()})
        rows = [{'a': 1}, {'a': 2}, {'a': 3}]
        status = t.insert(rows)
        assert status.num_rows == len(rows)
        assert status.num_excs == 0
        s1 = cl.create_view('s1', t, is_snapshot=True)
        v1 = cl.create_view('v1', s1, is_snapshot=False)
        s2 = cl.create_view('s2', v1, is_snapshot=True)
        v2 = cl.create_view('v2', s2, is_snapshot=False)

        def verify(s1: pxt.Table, s2: pxt.Table, v1: pxt.Table, v2: pxt.Table) -> None:
            assert s1.count() == len(rows)
            assert v1.count() == len(rows)
            assert s2.count() == len(rows)
            assert v2.count() == len(rows)

        verify(s1, s2, v1, v2)

        status = t.insert(rows)
        assert status.num_rows == len(rows)
        assert status.num_excs == 0
        verify(s1, s2, v1, v2)

        cl = pxt.Client(reload=True)
        s1 = cl.get_table('s1')
        s2 = cl.get_table('s2')
        v1 = cl.get_table('v1')
        v2 = cl.get_table('v2')
        verify(s1, s2, v1, v2)

    def test_snapshot_of_view_chain(self, test_client: pxt.Client) -> None:
        cl = test_client
        t = cl.create_table('tbl', {'a': IntType()})
        rows = [{'a': 1}, {'a': 2}, {'a': 3}]
        status = t.insert(rows)
        assert status.num_rows == len(rows)
        assert status.num_excs == 0
        v1 = cl.create_view('v1', t, is_snapshot=False)
        v2 = cl.create_view('v2', v1, is_snapshot=False)
        s = cl.create_view('s', v2, is_snapshot=True)

        def verify(v1: pxt.Table, v2: pxt.Table, s: pxt.Table) -> None:
            assert v1.count() == t.count()
            assert v2.count() == t.count()
            assert s.count() == len(rows)

        verify(v1, v2, s)

        status = t.insert(rows)
        assert status.num_rows == len(rows) * 3  # we also updated 2 views
        assert status.num_excs == 0
        verify(v1, v2, s)

        cl = pxt.Client(reload=True)
        v1 = cl.get_table('v1')
        v2 = cl.get_table('v2')
        s = cl.get_table('s')
        verify(v1, v2, s)

    def test_multiple_snapshot_paths(self, test_client: pxt.Client) -> None:
        cl = test_client
        t = create_test_tbl(cl)
        c4 = t.select(t.c4).order_by(t.c2).collect().to_pandas()['c4']
        orig_c3 = t.select(t.c3).collect().to_pandas()['c3']
        v = cl.create_view('v', base=t, schema={'v1': t.c3 + 1})
        s1 = cl.create_view('s1', v, is_snapshot=True)
        t.drop_column('c4')
        # s2 references the same view version as s1, but a different version of t (due to a schema change)
        s2 = cl.create_view('s2', v, is_snapshot=True)
        t.update({'c6': {'a': 17}})
        # s3 references the same view version as s2, but a different version of t (due to a data change)
        s3 = cl.create_view('s3', v, is_snapshot=True)
        t.update({'c3': t.c3 + 1})
        # s4 references different versions of t and v
        s4 = cl.create_view('s4', v, is_snapshot=True)

        def validate(t: pxt.Table, v: pxt.Table, s1: pxt.Table, s2: pxt.Table, s3: pxt.Table, s4: pxt.Table) -> None:
            # c4 is only visible in s1
            assert np.all(s1.select(s1.c4).collect().to_pandas()['c4'] == c4)
            with pytest.raises(AttributeError):
                _ = t.select(t.c4).collect()
            with pytest.raises(AttributeError):
                _ = v.select(v.c4).collect()
            with pytest.raises(AttributeError):
                _ = s2.select(s2.c4).collect()
            with pytest.raises(AttributeError):
                _ = s3.select(s3.c4).collect()
            with pytest.raises(AttributeError):
                _ = s4.select(s4.c4).collect()

            # c3
            assert np.all(t.select(t.c3).order_by(t.c2).collect().to_pandas()['c3'] == orig_c3 + 1)
            assert np.all(s1.select(s1.c3).order_by(s1.c2).collect().to_pandas()['c3'] == orig_c3)
            assert np.all(s2.select(s2.c3).order_by(s2.c2).collect().to_pandas()['c3'] == orig_c3)
            assert np.all(s3.select(s3.c3).order_by(s3.c2).collect().to_pandas()['c3'] == orig_c3)
            assert np.all(s4.select(s4.c3).order_by(s4.c2).collect().to_pandas()['c3'] == orig_c3 + 1)

            # v1
            assert np.all(
                v.select(v.v1).order_by(v.c2).collect().to_pandas()['v1'] == \
                t.select(t.c3).order_by(t.c2).collect().to_pandas()['c3'] + 1)
            assert np.all(s1.select(s1.v1).order_by(s1.c2).collect().to_pandas()['v1'] == orig_c3 + 1)
            assert np.all(s2.select(s2.v1).order_by(s2.c2).collect().to_pandas()['v1'] == orig_c3 + 1)
            assert np.all(s3.select(s3.v1).order_by(s3.c2).collect().to_pandas()['v1'] == orig_c3 + 1)
            assert np.all(
                s4.select(s4.v1).order_by(s4.c2).collect().to_pandas()['v1'] == \
                t.select(t.c3).order_by(t.c2).collect().to_pandas()['c3'] + 1)

        validate(t, v, s1, s2, s3, s4)

        # make sure it works after metadata reload
        cl = pxt.Client(reload=True)
        t, v = cl.get_table('test_tbl'), cl.get_table('v')
        s1, s2, s3, s4 = cl.get_table('s1'), cl.get_table('s2'), cl.get_table('s3'), cl.get_table('s4')
        validate(t, v, s1, s2, s3, s4)



from typing import Optional, Any, Dict, List, Tuple
import inspect
import logging
import sys

import numpy as np

from .signature import Signature, Parameter
from .batched_function import BatchedFunction
import pixeltable.env as env
import pixeltable.type_system as ts
import pixeltable.exceptions as excs


_logger = logging.getLogger('pixeltable')

class NOSFunction(BatchedFunction):
    def __init__(self, model_spec: 'nos.common.ModelSpec', self_path: str):
        return_type, param_types = self._convert_nos_signature(model_spec.signature)
        param_names = list(model_spec.signature.get_inputs_spec().keys())
        params = [
            Parameter(name, col_type, inspect.Parameter.POSITIONAL_OR_KEYWORD, is_batched=False)
            for name, col_type in zip(param_names, param_types)
        ]
        signature = Signature(return_type, params)

        # construct inspect.Signature
        py_params = [
            inspect.Parameter(name, inspect.Parameter.POSITIONAL_OR_KEYWORD)
            for name, col_type in zip(param_names, param_types)
        ]
        py_signature = inspect.Signature(py_params)
        super().__init__(signature, py_signature=py_signature, self_path=self_path)

        self.model_spec = model_spec
        self.nos_param_names = model_spec.signature.get_inputs_spec().keys()
        self.scalar_nos_param_names = []

        # for models on images
        self.img_param_pos: Optional[int] = None  # position of the image parameter in the function signature
        # for multi-resolution image models
        import nos
        self.img_batch_params: List[nos.common.ObjectTypeInfo] = []
        self.img_resolutions: List[int] = []  # for multi-resolution models
        self.batch_size: Optional[int] = None
        self.img_size: Optional[Tuple[int, int]] = None  # W, H

        # try to determine batch_size and img_size
        batch_size = sys.maxsize
        for pos, (param_name, type_info) in enumerate(model_spec.signature.get_inputs_spec().items()):
            if isinstance(type_info, list):
                assert isinstance(type_info[0].base_spec(), nos.common.ImageSpec)
                # this is a multi-resolution image model
                self.img_batch_params = type_info
                self.img_param_pos = pos
                self.img_resolutions = [
                    info.base_spec().shape[0] * info.base_spec().shape[1] for info in self.img_batch_params
                ]
            else:
                if not type_info.is_batched():
                    self.scalar_nos_param_names.append(param_name)
                else:
                    batch_size = min(batch_size, type_info.batch_size())

                if isinstance(type_info.base_spec(), nos.common.ImageSpec):
                    # this is a single-resolution image model
                    if type_info.base_spec().shape is not None:
                        self.img_size = (type_info.base_spec().shape[1], type_info.base_spec().shape[0])
                    self.img_param_pos = pos

        if batch_size != sys.maxsize:
            self.batch_size = batch_size

    def _convert_nos_type(
            self, type_info: 'nos.common.spec.ObjectTypeInfo', ignore_shape: bool = False
    ) -> ts.ColumnType:
        """Convert ObjectTypeInfo to ColumnType"""
        import nos
        if type_info.base_spec() is None:
            if type_info.base_type() == str:
                return ts.StringType()
            if type_info.base_type() == int:
                return ts.IntType()
            if type_info.base_type() == float:
                return ts.FloatType()
            if type_info.base_type() == bool:
                return ts.BoolType()
            else:
                raise excs.Error(f'Cannot convert {type_info} to ColumnType')
        elif isinstance(type_info.base_spec(), nos.common.ImageSpec):
            size = None
            if not ignore_shape and type_info.base_spec().shape is not None:
                size = (type_info.base_spec().shape[1], type_info.base_spec().shape[0])
            # TODO: set mode
            return ts.ImageType(size=size)
        elif isinstance(type_info.base_spec(), nos.common.TensorSpec):
            return ts.ArrayType(shape=type_info.base_spec().shape, dtype=ts.FloatType())
        else:
            raise excs.Error(f'Cannot convert {type_info} to ColumnType')

    def _convert_nos_signature(
            self, sig: 'nos.common.spec.FunctionSignature') -> Tuple[ts.ColumnType, List[ts.ColumnType]]:
        if len(sig.get_outputs_spec()) > 1:
            return_type = ts.JsonType()
        else:
            return_type = self._convert_nos_type(list(sig.get_outputs_spec().values())[0])
        param_types: List[ts.ColumnType] = []
        for _, type_info in sig.get_inputs_spec().items():
            # if there are multiple input shapes we leave them out of the ColumnType and deal with them in FunctionCall
            if isinstance(type_info, list):
                param_types.append(self._convert_nos_type(type_info[0], ignore_shape=True))
            else:
                param_types.append(self._convert_nos_type(type_info, ignore_shape=False))
        return return_type, param_types

    def is_multi_res_model(self) -> bool:
        return self.img_param_pos is not None and len(self.img_batch_params) > 0

    def get_batch_size(self, *args: Any, **kwargs: Any) -> Optional[int]:
        if self.batch_size is not None or len(self.img_batch_params) == 0 or len(args) == 0:
            return self.batch_size

        # return batch size appropriate for the given image size
        img_arg = args[self.img_param_pos]
        input_res = img_arg.size[0] * img_arg.size[1]
        batch_size, _ = self._select_model_res(input_res)
        return batch_size

    def _select_model_res(self, input_res: int) -> Tuple[int, Tuple[int, int]]:
        """Select the model resolution that is closest to the input resolution
        Returns: batch size, image size
        """
        deltas = [abs(res - input_res) for res in self.img_resolutions]
        idx = deltas.index(min(deltas))
        type_info = self.img_batch_params[idx]
        return type_info.batch_size(), (type_info.base_spec().shape[1], type_info.base_spec().shape[0])

    def invoke(self, arg_batches: List[List[Any]], kwarg_batches: Dict[str, List[Any]]) -> List[Any]:
        # check that scalar args are constant

        num_batch_rows = len(arg_batches[0])
        # if we need to rescale image args, and we're doing object detection, we need to rescale the
        # bounding boxes as well
        scale_factors = np.ndarray((num_batch_rows, 2), dtype=np.float32)

        target_res: Optional[Tuple[int, int]] = None
        if self.img_param_pos is not None:
            # for now, NOS will only receive RGB images
            arg_batches[self.img_param_pos] = \
                [img.convert('RGB') if img.mode != 'RGB' else img for img in arg_batches[self.img_param_pos ]]
            if self.is_multi_res_model():
                # we need to select the resolution that is closest to the input resolution
                sample_img = arg_batches[self.img_param_pos][0]
                _, target_res = self._select_model_res(sample_img.size[0] * sample_img.size[1])
            else:
                target_res = self.img_size

            if target_res is not None:
                # we need to record the scale factors and resize the images;
                # keep in mind that every image could have a different resolution
                scale_factors[:, 0] = \
                    [img.size[0] / target_res[0] for img in arg_batches[self.img_param_pos]]
                scale_factors[:, 1] = \
                    [img.size[1] / target_res[1] for img in arg_batches[self.img_param_pos]]
                arg_batches[self.img_param_pos] = [
                    # only resize if necessary
                    img.resize(target_res) if img.size != target_res else img
                    for img in arg_batches[self.img_param_pos]
                ]

        kwargs = {param_name: args for param_name, args in zip(self.nos_param_names, arg_batches)}
        # fix up scalar parameters
        kwargs.update(
            {param_name: kwargs[param_name][0] for param_name in self.scalar_nos_param_names})
        _logger.debug(
            f'Running NOS task {self.model_spec.task}: '
            f'batch_size={num_batch_rows} target_res={target_res}')
        result = env.Env.get().nos_client.Run(
            task=self.model_spec.task, model_name=self.model_spec.name, **kwargs)

        import nos
        if self.model_spec.task == nos.common.TaskType.OBJECT_DETECTION_2D and target_res is not None:
            # we need to rescale the bounding boxes
            result_bboxes = []  # workaround: result['bboxes'][*] is immutable
            for i, bboxes in enumerate(result['bboxes']):
                bboxes = np.copy(bboxes)
                nos_batch_row_idx = i
                bboxes[:, 0] *= scale_factors[nos_batch_row_idx, 0]
                bboxes[:, 1] *= scale_factors[nos_batch_row_idx, 1]
                bboxes[:, 2] *= scale_factors[nos_batch_row_idx, 0]
                bboxes[:, 3] *= scale_factors[nos_batch_row_idx, 1]
                result_bboxes.append(bboxes)
            result['bboxes'] = result_bboxes

        if len(result) == 1:
            key = list(result.keys())[0]
            row_results = result[key]
        else:
            # we rearrange result into one dict per row
            row_results = [{k: v[i].tolist() for k, v in result.items()} for i in range(num_batch_rows)]
        return row_results




from __future__ import annotations
from typing import Optional, List, Any, Dict, Tuple
import copy

import sqlalchemy as sql

from .expr import Expr
from .globals import print_slice
from .data_row import DataRow
from .row_builder import RowBuilder
import pixeltable.catalog as catalog


class ArraySlice(Expr):
    """
    Slice operation on an array, eg, t.array_col[:, 1:2].
    """
    def __init__(self, arr: Expr, index: Tuple):
        assert arr.col_type.is_array_type()
        # determine result type
        super().__init__(arr.col_type)
        self.components = [arr]
        self.index = index
        self.id = self._create_id()

    def __str__(self) -> str:
        index_strs: List[str] = []
        for el in self.index:
            if isinstance(el, int):
                index_strs.append(str(el))
            if isinstance(el, slice):
                index_strs.append(print_slice(el))
        return f'{self._array}[{", ".join(index_strs)}]'

    @property
    def _array(self) -> Expr:
        return self.components[0]

    def _equals(self, other: ArraySlice) -> bool:
        return self.index == other.index

    def _id_attrs(self) -> List[Tuple[str, Any]]:
        return super()._id_attrs() + [('index', self.index)]

    def sql_expr(self) -> Optional[sql.ClauseElement]:
        return None

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        val = data_row[self._array.slot_idx]
        data_row[self.slot_idx] = val[self.index]

    def _as_dict(self) -> Dict:
        index = []
        for el in self.index:
            if isinstance(el, slice):
                index.append([el.start, el.stop, el.step])
            else:
                index.append(el)
        return {'index': index, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr]) -> Expr:
        assert 'index' in d
        index = []
        for el in d['index']:
            if isinstance(el, list):
                index.append(slice(el[0], el[1], el[2]))
            else:
                index.append(el)
        return cls(components[0], tuple(index))




from __future__ import annotations

import logging
from typing import Optional, Union, Callable, Set

import sqlalchemy as sql
from pgvector.sqlalchemy import Vector

from pixeltable import exceptions as excs
from pixeltable.metadata import schema
from pixeltable.type_system import ColumnType, StringType
from .globals import is_valid_identifier

_logger = logging.getLogger('pixeltable')

class Column:
    """Representation of a column in the schema of a Table/DataFrame.

    A Column contains all the metadata necessary for executing queries and updates against a particular version of a
    table/view.
    """
    def __init__(
            self, name: str, col_type: Optional[ColumnType] = None,
            computed_with: Optional[Union['Expr', Callable]] = None,
            primary_key: bool = False, stored: Optional[bool] = None,
            indexed: bool = False,
            # these parameters aren't set by users
            col_id: Optional[int] = None):
        """Column constructor.

        Args:
            name: column name
            col_type: column type; can be None if the type can be derived from ``computed_with``
            computed_with: a callable or an Expr object that computes the column value
            primary_key: if True, this column is part of the primary key
            stored: determines whether a computed column is present in the stored table or recomputed on demand
            indexed: if True, this column has a nearest neighbor index (only valid for image columns)
            col_id: column ID (only used internally)

        Computed columns: those have a non-None ``computed_with`` argument

        - when constructed by the user: ``computed_with`` was constructed explicitly and is passed in;
          col_type is None
        - when loaded from md store: ``computed_with`` is set and col_type is set

        ``computed_with`` is a Callable:

        - the callable's parameter names must correspond to existing columns in the table for which this Column
          is being used
        - ``col_type`` needs to be set to the callable's return type

        ``stored`` (only valid for computed image columns):

        - if True: the column is present in the stored table
        - if False: the column is not present in the stored table and recomputed during a query
        - if None: the system chooses for you (at present, this is always False, but this may change in the future)

        indexed: only valid for image columns; if true, maintains an NN index for this column
        """
        if not is_valid_identifier(name):
            raise excs.Error(f"Invalid column name: '{name}'")
        self.name = name
        if col_type is None and computed_with is None:
            raise excs.Error(f'Column {name}: col_type is required if computed_with is not specified')

        self.value_expr: Optional['Expr'] = None
        self.compute_func: Optional[Callable] = None
        from pixeltable import exprs
        if computed_with is not None:
            value_expr = exprs.Expr.from_object(computed_with)
            if value_expr is None:
                # computed_with needs to be a Callable
                if not isinstance(computed_with, Callable):
                    raise excs.Error(
                        f'Column {name}: computed_with needs to be either a Pixeltable expression or a Callable, '
                        f'but it is a {type(computed_with)}')
                if col_type is None:
                    raise excs.Error(f'Column {name}: col_type is required if computed_with is a Callable')
                # we need to turn the computed_with function into an Expr, but this requires resolving
                # column name references and for that we need to wait until we're assigned to a Table
                self.compute_func = computed_with
            else:
                self.value_expr = value_expr.copy()
                self.col_type = self.value_expr.col_type

        if col_type is not None:
            self.col_type = col_type
        assert self.col_type is not None

        self.stored = stored
        self.dependent_cols: Set[Column] = set()  # cols with value_exprs that reference us; set by TableVersion
        self.id = col_id
        self.primary_key = primary_key

        # column in the stored table for the values of this Column
        self.sa_col: Optional[sql.schema.Column] = None

        # computed cols also have storage columns for the exception string and type
        self.sa_errormsg_col: Optional[sql.schema.Column] = None
        self.sa_errortype_col: Optional[sql.schema.Column] = None
        # indexed columns also have a column for the embeddings
        self.sa_idx_col: Optional[sql.schema.Column] = None
        from .table_version import TableVersion
        self.tbl: Optional[TableVersion] = None  # set by owning TableVersion

        if indexed and not self.col_type.is_image_type():
            raise excs.Error(f'Column {name}: indexed=True requires ImageType')
        self.is_indexed = indexed

    @classmethod
    def from_md(cls, col_id: int, md: schema.SchemaColumn, tbl: 'TableVersion') -> Column:
        """Construct a Column from metadata.

        Leaves out value_expr, because that requires TableVersion.cols to be complete.
        """
        col = cls(
            md.name, col_type=ColumnType.from_dict(md.col_type), primary_key=md.is_pk,
            stored=md.stored, indexed=md.is_indexed, col_id=col_id)
        col.tbl = tbl
        return col

    def __hash__(self) -> int:
        assert self.tbl is not None
        return hash((self.tbl.id, self.id))

    def check_value_expr(self) -> None:
        assert self.value_expr is not None
        if self.stored == False and self.is_computed and self.has_window_fn_call():
            raise excs.Error(
                f'Column {self.name}: stored={self.stored} not supported for columns computed with window functions:'
                f'\n{self.value_expr}')

    def has_window_fn_call(self) -> bool:
        if self.value_expr is None:
            return False
        from pixeltable import exprs
        l = list(self.value_expr.subexprs(filter=lambda e: isinstance(e, exprs.FunctionCall) and e.is_window_fn_call))
        return len(l) > 0

    @property
    def is_computed(self) -> bool:
        return self.compute_func is not None or self.value_expr is not None

    @property
    def is_stored(self) -> bool:
        """Returns True if column is materialized in the stored table."""
        assert self.stored is not None
        return self.stored

    @property
    def records_errors(self) -> bool:
        """True if this column also stores error information."""
        return self.is_stored and (self.is_computed or self.col_type.is_media_type())

    def source(self) -> None:
        """
        If this is a computed col and the top-level expr is a function call, print the source, if possible.
        """
        from pixeltable import exprs
        if self.value_expr is None or not isinstance(self.value_expr, exprs.FunctionCall):
            return
        self.value_expr.fn.source()

    def create_sa_cols(self) -> None:
        """
        These need to be recreated for every new table schema version.
        """
        assert self.is_stored
        # all storage columns are nullable (we deal with null errors in Pixeltable directly)
        self.sa_col = sql.Column(self.storage_name(), self.col_type.to_sa_type(), nullable=True)
        if self.is_computed or self.col_type.is_media_type():
            self.sa_errormsg_col = sql.Column(self.errormsg_storage_name(), StringType().to_sa_type(), nullable=True)
            self.sa_errortype_col = sql.Column(self.errortype_storage_name(), StringType().to_sa_type(), nullable=True)
        if self.is_indexed:
            self.sa_idx_col = sql.Column(self.index_storage_name(), Vector(512), nullable=True)

    def storage_name(self) -> str:
        assert self.id is not None
        assert self.is_stored
        return f'col_{self.id}'

    def errormsg_storage_name(self) -> str:
        return f'{self.storage_name()}_errormsg'

    def errortype_storage_name(self) -> str:
        return f'{self.storage_name()}_errortype'

    def index_storage_name(self) -> str:
        return f'{self.storage_name()}_idx_0'

    def __str__(self) -> str:
        return f'{self.name}: {self.col_type}'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Column):
            return False
        assert self.tbl is not None
        assert other.tbl is not None
        return self.tbl.id == other.tbl.id and self.id == other.id




from typing import Dict, Any, List, Tuple, Generator, Optional, Iterable
import logging
import dataclasses
import enum

from .base import ComponentIterator

from pixeltable.type_system import ColumnType, DocumentType, StringType, IntType, JsonType
from pixeltable.exceptions import Error
from pixeltable.env import Env
from pixeltable.utils.documents import get_document_handle


_logger = logging.getLogger('pixeltable')


class ChunkMetadata(enum.Enum):
    TITLE = 1
    HEADINGS = 2
    SOURCELINE = 3


class Separator(enum.Enum):
    HEADING = 1
    PARAGRAPH = 2
    SENTENCE = 3
    TOKEN_LIMIT = 4
    CHAR_LIMIT = 5


@dataclasses.dataclass
class DocumentSectionMd:
    """Metadata for a subsection of a document (ie, a structural element like a heading or paragraph)"""
    source_line: int

    # the stack of headings up to the most recently observed one;
    # eg, if the most recent one was an h2, 'headings' would contain keys 1 and 2, but nothing below that
    headings: Dict[int, str]


@dataclasses.dataclass
class DocumentSection:
    """A single document chunk, according to some of the splitting criteria"""
    text: Optional[str]
    md: Optional[DocumentSectionMd]


class DocumentSplitter(ComponentIterator):
    """"Iterator over pieces of a document"""
    MD_COLUMN_TYPES = {
        ChunkMetadata.TITLE: StringType(),
        ChunkMetadata.HEADINGS: JsonType(),
        ChunkMetadata.SOURCELINE: IntType()
    }

    def __init__(
            self, document: str, *, separators: str, limit: int = 0, overlap: int = 0, metadata: str = '',
            html_skip_tags: List[str] = None, tiktoken_encoding: Optional[str] = 'cl100k_base',
            tiktoken_target_model: Optional[str] = None
    ):
        import bs4
        if html_skip_tags is None:
            html_skip_tags = ['nav']
        with open(document, 'r') as fh:
            s = fh.read()
            self._doc_handle = get_document_handle(s)
            assert self._doc_handle is not None
        self._separators = [Separator[s.upper()] for s in separators.split(',')]
        self._md_fields = [ChunkMetadata[m.upper()] for m in metadata.split(',')] if len(metadata) > 0 else []
        self._doc_title = \
            self._doc_handle.bs_doc.title.get_text().strip() if self._doc_handle.bs_doc is not None else ''
        self._limit = limit
        self._skip_tags = html_skip_tags
        self._overlap = overlap
        self._tiktoken_encoding = tiktoken_encoding
        self._tiktoken_target_model = tiktoken_target_model

        # set up processing pipeline
        if self._doc_handle.format == DocumentType.DocumentFormat.HTML:
            assert self._doc_handle.bs_doc is not None
            self._sections = self._html_sections()
        else:
            assert self._doc_handle.md_ast is not None
            self._sections = self._markdown_sections()
        if Separator.SENTENCE in self._separators:
            self._sections = self._sentence_sections(self._sections)
        if Separator.TOKEN_LIMIT in self._separators:
            self._sections = self._token_chunks(self._sections)
        if Separator.CHAR_LIMIT in self._separators:
            self._sections = self._char_chunks(self._sections)

    @classmethod
    def input_schema(cls) -> Dict[str, ColumnType]:
        return {
            'document': DocumentType(nullable=False),
            'separators': StringType(nullable=False),
            'metadata': StringType(nullable=True),
            'limit': IntType(nullable=True),
            'overlap': IntType(nullable=True),
            'skip_tags': StringType(nullable=True),
            'tiktoken_encoding': StringType(nullable=True),
            'tiktoken_target_model': StringType(nullable=True),
        }

    @classmethod
    def output_schema(cls, *args: Any, **kwargs: Any) -> Tuple[Dict[str, ColumnType], List[str]]:
        schema = {'text': StringType()}
        if 'metadata' in kwargs and len(kwargs['metadata']) > 0:
            md_fields = kwargs['metadata'].split(',')
            for md_field in md_fields:
                if not hasattr(ChunkMetadata, md_field.upper()):
                    raise Error(f'Invalid metadata field {md_field}')
                schema[md_field.lower()] = cls.MD_COLUMN_TYPES[ChunkMetadata[md_field.upper()]]

        assert 'separators' in kwargs
        separators = kwargs['separators'].split(',')
        for separator in separators:
            if not hasattr(Separator, separator.upper()):
                raise Error(f'Invalid separator {separator}')

        # check dependencies
        if 'sentence' in separators:
            Env.get().require_package('spacy')
        if 'token_limit' in separators:
            Env.get().require_package('tiktoken')

        if 'limit' in kwargs or 'overlap' in kwargs:
            if 'token_limit' not in separators and 'char_limit' not in separators:
                raise Error('limit/overlap requires the "token_limit" or "char_limit" separator')
            if 'limit' in kwargs and int(kwargs['limit']) <= 0:
                raise Error('"limit" must be an integer > 0')
            if 'overlap' in kwargs and int(kwargs['overlap']) < 0:
                raise Error('"overlap" must be an integer >= 0')
        if 'token_limit' in separators or 'char_limit' in separators:
            if 'token_limit' in separators and 'char_limit' in separators:
                raise Error('Cannot specify both "token_limit" and "char_limit" separators')
            if 'limit' not in kwargs:
                raise Error('limit is required with "token_limit"/"char_limit" separators')

        return schema, []

    def __next__(self) -> Dict[str, Any]:
        while True:
            section = next(self._sections)
            if section.text is None:
                continue
            result = {'text': section.text}
            for md_field in self._md_fields:
                if md_field == ChunkMetadata.TITLE:
                    result[md_field.name.lower()] = self._doc_title
                elif md_field == ChunkMetadata.HEADINGS:
                    result[md_field.name.lower()] = section.md.headings
                elif md_field == ChunkMetadata.SOURCELINE:
                    result[md_field.name.lower()] = section.md.source_line
            return result

    def _html_sections(self) -> Generator[DocumentSection, None, None]:
        """Create DocumentSections reflecting the html-specific separators"""
        import bs4
        emit_on_paragraph = Separator.PARAGRAPH in self._separators or Separator.SENTENCE in self._separators
        emit_on_heading = Separator.HEADING in self._separators or emit_on_paragraph
        # current state
        text_section = ''  # currently accumulated text
        headings: Dict[int, str] = {}   # current state of observed headings (level -> text)
        sourceline = 0  # most recently seen sourceline

        def update_md(el: bs4.Tag) -> None:
            # update current state
            nonlocal headings, sourceline
            sourceline = el.sourceline
            if el.name in ['h1', 'h2', 'h3', 'h4', 'h5', 'h6']:
                level = int(el.name[1])
                # remove the previously seen lower levels
                lower_levels = [l for l in headings.keys() if l > level]
                for l in lower_levels:
                    del headings[l]
                headings[level] = el.get_text().strip()

        def emit() -> None:
            nonlocal text_section, headings, sourceline
            if len(text_section) > 0:
                md = DocumentSectionMd(sourceline, headings.copy())
                yield DocumentSection(text=text_section, md=md)
                text_section = ''

        def process_element(el: bs4.PageElement) -> Generator[DocumentSection, None, None]:
            # process the element and emit sections as necessary
            nonlocal text_section, headings, sourceline, emit_on_heading, emit_on_paragraph
            if el.name in self._skip_tags:
                return

            if isinstance(el, bs4.NavigableString):
                # accumulate text until we see a tag we care about
                text = el.get_text().strip()
                if len(text) > 0:
                    text_section += ' ' + text
                return

            if el.name in ['h1', 'h2', 'h3', 'h4', 'h5', 'h6']:
                if emit_on_heading:
                    yield from emit()
                update_md(el)
            elif el.name == 'p':
                if emit_on_paragraph:
                    yield from emit()
                update_md(el)
            for child in el.children:
                yield from process_element(child)

        yield from process_element(self._doc_handle.bs_doc)
        yield from emit()

    def _markdown_sections(self) -> Generator[DocumentSection, None, None]:
        """Create DocumentSections reflecting the html-specific separators"""
        assert self._doc_handle.md_ast is not None
        emit_on_paragraph = Separator.PARAGRAPH in self._separators or Separator.SENTENCE in self._separators
        emit_on_heading = Separator.HEADING in self._separators or emit_on_paragraph
        # current state
        text_section = ''  # currently accumulated text
        headings: Dict[int, str] = {}   # current state of observed headings (level -> text)

        def update_headings(heading: Dict) -> None:
            # update current state
            nonlocal headings
            assert 'type' in heading and heading['type'] == 'heading'
            level = heading['attrs']['level']
            text = heading['children'][0]['raw'].strip()
            # remove the previously seen lower levels
            lower_levels = [l for l in headings.keys() if l > level]
            for l in lower_levels:
                del headings[l]
            headings[level] = text

        def emit() -> None:
            nonlocal text_section, headings
            if len(text_section) > 0:
                md = DocumentSectionMd(0, headings.copy())
                yield DocumentSection(text=text_section, md=md)
                text_section = ''

        def process_element(el: Dict) -> Generator[DocumentSection, None, None]:
            # process the element and emit sections as necessary
            nonlocal text_section, headings, emit_on_heading, emit_on_paragraph
            assert 'type' in el

            if el['type'] == 'text':
                # accumulate text until we see a separator element
                text = el['raw'].strip()
                if len(text) > 0:
                    text_section += ' ' + text
                return

            if el['type'] == 'heading':
                if emit_on_heading:
                    yield from emit()
                update_headings(el)
            elif el['type'] == 'paragraph':
                if emit_on_paragraph:
                    yield from emit()
            if 'children' not in el:
                return
            for child in el['children']:
                yield from process_element(child)

        for el in self._doc_handle.md_ast:
            yield from process_element(el)
        yield from emit()

    def _sentence_sections(self, input_sections: Iterable[DocumentSection]) -> Generator[DocumentSection, None, None]:
        """Split the input sections into sentences"""
        for section in input_sections:
            if section.text is not None:
                doc = Env.get().spacy_nlp(section.text)
                for sent in doc.sents:
                    yield DocumentSection(text=sent.text, md=section.md)

    def _token_chunks(self, input: Iterable[DocumentSection]) -> Generator[DocumentSection, None, None]:
        import tiktoken
        if self._tiktoken_target_model is not None:
            encoding = tiktoken.encoding_for_model(self._tiktoken_target_model)
        else:
            encoding = tiktoken.get_encoding(self._tiktoken_encoding)
        assert self._limit > 0 and self._overlap >= 0

        for section in input:
            if section.text is None:
                continue
            tokens = encoding.encode(section.text)
            start_idx = 0
            while start_idx < len(tokens):
                end_idx = min(start_idx + self._limit, len(tokens))
                text = encoding.decode(tokens[start_idx:end_idx])
                yield DocumentSection(text=text, md=section.md)
                start_idx += self._limit - self._overlap

    def _char_chunks(self, input: Iterable[DocumentSection]) -> Generator[DocumentSection, None, None]:
        for section in input:
            if section.text is None:
                continue
            start_idx = 0
            while start_idx < len(section.text):
                end_idx = min(start_idx + self._limit, len(section.text))
                text = section.text[start_idx:end_idx]
                yield DocumentSection(text=text, md=section.md)
                start_idx += self._limit - self._overlap

    def close(self) -> None:
        pass

    def set_pos(self, pos: int) -> None:
        pass



from typing import Tuple, List, Optional
import types
import sys

import pixeltable.func as func
import pixeltable.type_system as ts
import pixeltable.env as env


def create_nos_modules() -> List[types.ModuleType]:
    """Create module pixeltable.functions.nos with one submodule per task and return the submodules"""
    models = env.Env.get().nos_client.ListModels()
    model_info = [env.Env.get().nos_client.GetModelInfo(model) for model in models]
    model_info.sort(key=lambda info: info.task.value)

    module_name = 'pixeltable.functions.nos'
    nos_module = types.ModuleType(module_name)
    nos_module.__package__ = 'pixeltable.functions'
    sys.modules[module_name] = nos_module

    prev_task = ''
    new_modules: List[types.ModuleType] = []
    sub_module: Optional[types.ModuleType] = None
    for info in model_info:
        if info.task.value != prev_task:
            # we construct one submodule per task
            namespace = info.task.name.lower()
            submodule_name = f'{module_name}.{namespace}'
            sub_module = types.ModuleType(submodule_name)
            sub_module.__package__ = module_name
            setattr(nos_module, namespace, sub_module)
            new_modules.append(sub_module)
            sys.modules[submodule_name] = sub_module
            prev_task = info.task.value

        # add a Function for this model to the module
        model_id = info.name.replace("/", "_").replace("-", "_")
        pt_func = func.NOSFunction(info, f'{submodule_name}.{model_id}')
        setattr(sub_module, model_id, pt_func)

    return new_modules



from typing import List
import dataclasses
import logging


_logger = logging.getLogger('pixeltable')

# name of the position column in a component view
POS_COLUMN_NAME = 'pos'

@dataclasses.dataclass
class UpdateStatus:
    num_rows: int = 0
    # TODO: disambiguate what this means: # of slots computed or # of columns computed?
    num_computed_values: int = 0
    num_excs: int = 0
    updated_cols: List[str] = dataclasses.field(default_factory=list)
    cols_with_excs: List[str] = dataclasses.field(default_factory=list)

def is_valid_identifier(name: str) -> bool:
    return name.isidentifier() and not name.startswith('_')

def is_valid_path(path: str, empty_is_valid : bool) -> bool:
    if path == '':
        return empty_is_valid

    for part in path.split('.'):
        if not is_valid_identifier(part):
            return False
    return True

def is_system_column_name(name: str) -> bool:
    return name == POS_COLUMN_NAME



import glob
import logging
import os
import subprocess

import pgserver
import pytest

import pixeltable as pxt
from pixeltable.env import Env
from pixeltable.tests.conftest import clean_db

_logger = logging.getLogger('pixeltable')


class TestMigration:

    @pytest.mark.skip(reason='Suspended')
    def test_db_migration(self, init_env) -> None:
        env = Env.get()
        pg_package_dir = os.path.dirname(pgserver.__file__)
        pg_restore_binary = f'{pg_package_dir}/pginstall/bin/pg_restore'
        _logger.info(f'Using pg_restore binary at: {pg_restore_binary}')
        dump_files = glob.glob('pixeltable/tests/data/dbdumps/*.dump.gz')
        dump_files.sort()
        for dump_file in dump_files:
            _logger.info(f'Testing migration from DB dump {dump_file}.')
            _logger.info(f'DB URL: {env.db_url}')
            clean_db(restore_tables=False)
            with open(dump_file, 'rb') as dump:
                gunzip_process = subprocess.Popen(
                    ["gunzip", "-c"],
                    stdin=dump,
                    stdout=subprocess.PIPE
                )
                subprocess.run(
                    [pg_restore_binary, '-d', env.db_url, '-U', 'postgres'],
                    stdin=gunzip_process.stdout,
                    check=True
                )
            # TODO(aaron-siegel) This will test that the migration succeeds without raising any exceptions.
            # We should also add some assertions to sanity-check the outcome.
            _ = pxt.Client()



import logging
import os
from typing import Optional

import pixeltable as pxt
import pixeltable.exceptions as excs
from pixeltable import env


@pxt.udf
def chat_completions(
        prompt: str,
        model: str,
        *,
        max_tokens: Optional[int] = None,
        repetition_penalty: Optional[float] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        temperature: Optional[float] = None
) -> dict:
    initialize()
    kwargs = {
        'max_tokens': max_tokens,
        'repetition_penalty': repetition_penalty,
        'top_k': top_k,
        'top_p': top_p,
        'temperature': temperature
    }
    kwargs_not_none = dict(filter(lambda x: x[1] is not None, kwargs.items()))
    import fireworks.client
    return fireworks.client.Completion.create(
        model=model,
        prompt_or_messages=prompt,
        **kwargs_not_none
    ).dict()


def initialize():
    global _is_fireworks_initialized
    if _is_fireworks_initialized:
        return

    _logger.info('Initializing Fireworks client.')

    config = pxt.env.Env.get().config

    if 'fireworks' in config and 'api_key' in config['fireworks']:
        api_key = config['fireworks']['api_key']
    else:
        api_key = os.environ.get('FIREWORKS_API_KEY')
    if api_key is None or api_key == '':
        raise excs.Error('Fireworks client not initialized (no API key configured).')

    import fireworks.client

    fireworks.client.api_key = api_key
    _is_fireworks_initialized = True


_logger = logging.getLogger('pixeltable')
_is_fireworks_initialized = False



from typing import Dict, Any

import pytest

import pixeltable as pxt
from pixeltable import catalog
from pixeltable.env import Env
from pixeltable.functions.pil.image import blend
from pixeltable.iterators import FrameIterator
from pixeltable.tests.utils import get_video_files, skip_test_if_not_installed, get_sentences, get_image_files
from pixeltable.type_system import VideoType, StringType, JsonType, ImageType, BoolType, FloatType, ArrayType


class TestFunctions:
    def test_pil(self, img_tbl: catalog.Table) -> None:
        t = img_tbl
        _ = t[t.img, t.img.rotate(90), blend(t.img, t.img.rotate(90), 0.5)].show()

    def test_eval_detections(self, test_client: pxt.Client) -> None:
        skip_test_if_not_installed('nos')
        cl = test_client
        video_t = cl.create_table('video_tbl', {'video': VideoType()})
        # create frame view
        args = {'video': video_t.video, 'fps': 1}
        v = cl.create_view('test_view', video_t, iterator_class=FrameIterator, iterator_args=args)

        files = get_video_files()
        video_t.insert(video=files[-1])
        v.add_column(frame_s=v.frame.resize([640, 480]))
        from pixeltable.functions.nos.object_detection_2d import yolox_nano, yolox_small, yolox_large
        v.add_column(detections_a=yolox_nano(v.frame_s))
        v.add_column(detections_b=yolox_small(v.frame_s))
        v.add_column(gt=yolox_large(v.frame_s))
        from pixeltable.functions.eval import eval_detections, mean_ap
        res = v.select(
            eval_detections(
                v.detections_a.bboxes, v.detections_a.labels, v.detections_a.scores, v.gt.bboxes, v.gt.labels
            )).show()
        v.add_column(
            eval_a=eval_detections(
                v.detections_a.bboxes, v.detections_a.labels, v.detections_a.scores, v.gt.bboxes, v.gt.labels))
        v.add_column(
            eval_b=eval_detections(
                v.detections_b.bboxes, v.detections_b.labels, v.detections_b.scores, v.gt.bboxes, v.gt.labels))
        ap_a = v.select(mean_ap(v.eval_a)).show()[0, 0]
        ap_b = v.select(mean_ap(v.eval_b)).show()[0, 0]
        common_classes = set(ap_a.keys()) & set(ap_b.keys())

        ## TODO: following assertion is failing on CI, 
        # It is not necessarily a bug, as assert codition is not expected to be always true
        # for k in common_classes:
        # assert ap_a[k] <= ap_b[k]

    def test_str(self, test_client: pxt.Client) -> None:
        cl = test_client
        t = cl.create_table('test_tbl', {'input': StringType()})
        from pixeltable.functions.string import str_format
        t.add_column(s1=str_format('ABC {0}', t.input))
        t.add_column(s2=str_format('DEF {this}', this=t.input))
        t.add_column(s3=str_format('GHI {0} JKL {this}', t.input, this=t.input))
        status = t.insert(input='MNO')
        assert status.num_rows == 1
        assert status.num_excs == 0
        row = t.head()[0]
        assert row == {'input': 'MNO', 's1': 'ABC MNO', 's2': 'DEF MNO', 's3': 'GHI MNO JKL MNO'}

    def test_openai(self, test_client: pxt.Client) -> None:
        skip_test_if_not_installed('openai')
        TestFunctions.skip_test_if_no_openai_client()
        if Env.get().openai_client is None:
            pytest.skip(f'OpenAI client does not exist (missing API key?).')
        cl = test_client
        t = cl.create_table('test_tbl', {'input': StringType()})
        from pixeltable.functions.openai import chat_completions, embeddings, moderations
        msgs = [
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": t.input}
        ]
        t.add_column(input_msgs=msgs)
        t.add_column(chat_output=chat_completions(model='gpt-3.5-turbo', messages=t.input_msgs))
        # with inlined messages
        t.add_column(chat_output2=chat_completions(model='gpt-3.5-turbo', messages=msgs))
        t.add_column(ada_embed=embeddings(model='text-embedding-ada-002', input=t.input))
        t.add_column(text_3=embeddings(model='text-embedding-3-small', input=t.input))
        t.add_column(moderation=moderations(input=t.input))
        t.insert(input='I find you really annoying')
        _ = t.head()

    def test_gpt_4_vision(self, test_client: pxt.Client) -> None:
        skip_test_if_not_installed('openai')
        TestFunctions.skip_test_if_no_openai_client()
        cl = test_client
        t = cl.create_table('test_tbl', {'prompt': StringType(), 'img': ImageType()})
        from pixeltable.functions.openai import chat_completions
        from pixeltable.functions.string import str_format
        msgs = [
            {'role': 'user',
             'content': [
                 {'type': 'text', 'text': t.prompt},
                 {'type': 'image_url', 'image_url': {
                     'url': str_format('data:image/png;base64,{0}', t.img.b64_encode())
                 }}
             ]}
        ]
        t.add_column(response=chat_completions(model='gpt-4-vision-preview', messages=msgs, max_tokens=300))
        t.add_column(response_content=t.response.choices[0].message.content)
        t.insert(prompt="What's in this image?", img=_sample_image_url)
        result = t.collect()['response_content'][0]
        assert len(result) > 0

    @staticmethod
    def skip_test_if_no_openai_client() -> None:
        if Env.get().openai_client is None:
            pytest.skip(f'OpenAI client does not exist (missing API key?)')

    def test_together(self, test_client: pxt.Client) -> None:
        skip_test_if_not_installed('together')
        if not Env.get().has_together_client:
            pytest.skip(f'Together client does not exist (missing API key?)')
        cl = test_client
        t = cl.create_table('test_tbl', {'input': StringType()})
        from pixeltable.functions.together import completions
        t.add_column(output=completions(prompt=t.input, model='mistralai/Mixtral-8x7B-v0.1', stop=['\n']))
        t.add_column(output_text=t.output.output.choices[0].text)
        t.insert(input='I am going to the ')
        result = t.select(t.output_text).collect()['output_text'][0]
        assert len(result) > 0

    def test_fireworks(self, test_client: pxt.Client) -> None:
        skip_test_if_not_installed('fireworks')
        try:
            from pixeltable.functions.fireworks import initialize
            initialize()
        except:
            pytest.skip(f'Fireworks client does not exist (missing API key?)')
        cl = test_client
        t = cl.create_table('test_tbl', {'input': StringType()})
        from pixeltable.functions.fireworks import chat_completions
        t['output'] = chat_completions(prompt=t.input, model='accounts/fireworks/models/llama-v2-7b-chat', max_tokens=256).choices[0].text
        t.insert(input='I am going to the ')
        result = t.select(t.output).collect()['output'][0]
        assert len(result) > 0

    def test_hf_function(self, test_client: pxt.Client) -> None:
        skip_test_if_not_installed('sentence_transformers')
        cl = test_client
        t = cl.create_table('test_tbl', {'input': StringType(), 'bool_col': BoolType()})
        from pixeltable.functions.huggingface import sentence_transformer
        model_id = 'intfloat/e5-large-v2'
        t.add_column(e5=sentence_transformer(t.input, model_id=model_id))
        sents = get_sentences()
        status = t.insert({'input': s, 'bool_col': True} for s in sents)
        assert status.num_rows == len(sents)
        assert status.num_excs == 0

        # verify handling of constant params
        with pytest.raises(ValueError) as exc_info:
            t.add_column(e5_2=sentence_transformer(t.input, model_id=t.input))
        assert ': parameter model_id must be a constant value' in str(exc_info.value)
        with pytest.raises(ValueError) as exc_info:
            t.add_column(e5_2=sentence_transformer(t.input, model_id=model_id, normalize_embeddings=t.bool_col))
        assert ': parameter normalize_embeddings must be a constant value' in str(exc_info.value)

        # make sure this doesn't cause an exception
        # TODO: is there some way to capture the output?
        t.describe()

    def test_sentence_transformer(self, test_client: pxt.Client) -> None:
        skip_test_if_not_installed('sentence_transformers')
        cl = test_client
        t = cl.create_table('test_tbl', {'input': StringType(), 'input_list': JsonType()})
        sents = get_sentences(10)
        status = t.insert({'input': s, 'input_list': sents} for s in sents)
        assert status.num_rows == len(sents)
        assert status.num_excs == 0

        # run multiple models one at a time in order to exercise batching
        from pixeltable.functions.huggingface import sentence_transformer, sentence_transformer_list
        model_ids = ['sentence-transformers/all-mpnet-base-v2', 'BAAI/bge-reranker-base']
        num_dims = [768, 768]
        for idx, model_id in enumerate(model_ids):
            col_name = f'embed{idx}'
            t[col_name] = sentence_transformer(t.input, model_id=model_id, normalize_embeddings=True)
            assert t.column_types()[col_name] == ArrayType((None,), dtype=FloatType(), nullable=False)
            list_col_name = f'embed_list{idx}'
            t[list_col_name] = sentence_transformer_list(t.input_list, model_id=model_id, normalize_embeddings=True)
            assert t.column_types()[list_col_name] == JsonType()

        def verify_row(row: Dict[str, Any]) -> None:
            for idx, (_, d) in enumerate(zip(model_ids, num_dims)):
                assert row[f'embed{idx}'].shape == (d,)
                assert len(row[f'embed_list{idx}']) == len(sents)
                assert all(len(v) == d for v in row[f'embed_list{idx}'])

        verify_row(t.tail(1)[0])

        # execution still works after reload
        cl = pxt.Client(reload=True)
        t = cl.get_table('test_tbl')
        status = t.insert({'input': s, 'input_list': sents} for s in sents)
        assert status.num_rows == len(sents)
        assert status.num_excs == 0
        verify_row(t.tail(1)[0])

    def test_cross_encoder(self, test_client: pxt.Client) -> None:
        skip_test_if_not_installed('sentence_transformers')
        cl = test_client
        t = cl.create_table('test_tbl', {'input': StringType(), 'input_list': JsonType()})
        sents = get_sentences(10)
        status = t.insert({'input': s, 'input_list': sents} for s in sents)
        assert status.num_rows == len(sents)
        assert status.num_excs == 0

        # run multiple models one at a time in order to exercise batching
        from pixeltable.functions.huggingface import cross_encoder, cross_encoder_list
        model_ids = ['cross-encoder/ms-marco-MiniLM-L-6-v2', 'cross-encoder/ms-marco-TinyBERT-L-2-v2']
        for idx, model_id in enumerate(model_ids):
            col_name = f'embed{idx}'
            t[col_name] = cross_encoder(t.input, t.input, model_id=model_id)
            assert t.column_types()[col_name] == FloatType()
            list_col_name = f'embed_list{idx}'
            t[list_col_name] = cross_encoder_list(t.input, t.input_list, model_id=model_id)
            assert t.column_types()[list_col_name] == JsonType()

        def verify_row(row: Dict[str, Any]) -> None:
            for i in range(len(model_ids)):
                assert len(row[f'embed_list{idx}']) == len(sents)
                assert all(isinstance(v, float) for v in row[f'embed_list{idx}'])

        verify_row(t.tail(1)[0])

        # execution still works after reload
        cl = pxt.Client(reload=True)
        t = cl.get_table('test_tbl')
        status = t.insert({'input': s, 'input_list': sents} for s in sents)
        assert status.num_rows == len(sents)
        assert status.num_excs == 0
        verify_row(t.tail(1)[0])

    def test_clip(self, test_client: pxt.Client) -> None:
        skip_test_if_not_installed('transformers')
        cl = test_client
        t = cl.create_table('test_tbl', {'text': StringType(), 'img': ImageType()})
        num_rows = 10
        sents = get_sentences(num_rows)
        imgs = get_image_files()[:num_rows]
        status = t.insert({'text': text, 'img': img} for text, img in zip(sents, imgs))
        assert status.num_rows == len(sents)
        assert status.num_excs == 0

        # run multiple models one at a time in order to exercise batching
        from pixeltable.functions.huggingface import clip_text, clip_image
        model_ids = ['openai/clip-vit-base-patch32', 'laion/CLIP-ViT-B-32-laion2B-s34B-b79K']
        for idx, model_id in enumerate(model_ids):
            col_name = f'embed_text{idx}'
            t[col_name] = clip_text(t.text, model_id=model_id)
            assert t.column_types()[col_name] == ArrayType((None,), dtype=FloatType(), nullable=False)
            col_name = f'embed_img{idx}'
            t[col_name] = clip_image(t.img, model_id=model_id)
            assert t.column_types()[col_name] == ArrayType((None,), dtype=FloatType(), nullable=False)

        def verify_row(row: Dict[str, Any]) -> None:
            for idx, _ in enumerate(model_ids):
                assert row[f'embed_text{idx}'].shape == (512,)
                assert row[f'embed_img{idx}'].shape == (512,)

        verify_row(t.tail(1)[0])

        # execution still works after reload
        cl = pxt.Client(reload=True)
        t = cl.get_table('test_tbl')
        status = t.insert({'text': text, 'img': img} for text, img in zip(sents, imgs))
        assert status.num_rows == len(sents)
        assert status.num_excs == 0
        verify_row(t.tail(1)[0])

    def test_detr_for_object_detection(self, test_client: pxt.Client) -> None:
        skip_test_if_not_installed('transformers')
        cl = test_client
        t = cl.create_table('test_tbl', {'img': ImageType()})
        from pixeltable.functions.huggingface import detr_for_object_detection
        t['detect'] = detr_for_object_detection(t.img, model_id='facebook/detr-resnet-50', threshold=0.8)
        status = t.insert(img=_sample_image_url)
        assert status.num_rows == 1
        assert status.num_excs == 0
        result = t.select(t.detect).collect()[0]['detect']
        assert 'orange' in result['label_text']
        assert 'bowl' in result['label_text']
        assert 'broccoli' in result['label_text']


_sample_image_url = \
    'https://raw.githubusercontent.com/pixeltable/pixeltable/master/docs/source/data/images/000000000009.jpg'



from typing import Any


def get_client() -> Any:
    import boto3
    import botocore
    try:
        boto3.Session().get_credentials().get_frozen_credentials()
        return boto3.client('s3')  # credentials are available
    except AttributeError:
        # No credentials available, use unsigned mode
        config = botocore.config.Config(signature_version=botocore.UNSIGNED)
        return boto3.client('s3', config=config)


from __future__ import annotations
from typing import Optional, List, Any, Dict, Tuple
from uuid import UUID

import sqlalchemy as sql

from .expr import Expr
from .data_row import DataRow
from .row_builder import RowBuilder
import pixeltable.iterators as iters
import pixeltable.exceptions as excs
import pixeltable.catalog as catalog


class ColumnRef(Expr):
    """A reference to a table column

    When this reference is created in the context of a view, it can also refer to a column of the view base.
    For that reason, a ColumnRef needs to be serialized with the qualifying table id (column ids are only
    unique in the context of a particular table).
    """
    def __init__(self, col: catalog.Column):
        super().__init__(col.col_type)
        assert col.tbl is not None
        self.col = col
        self.is_unstored_iter_col = \
            col.tbl.is_component_view() and col.tbl.is_iterator_column(col) and not col.is_stored
        self.iter_arg_ctx: Optional[RowBuilder.EvalCtx] = None
        # number of rowid columns in the base table
        self.base_rowid_len = col.tbl.base.num_rowid_columns() if self.is_unstored_iter_col else 0
        self.base_rowid = [None] * self.base_rowid_len
        self.iterator: Optional[iters.ComponentIterator] = None
        # index of the position column in the view's primary key; don't try to reference tbl.store_tbl here
        self.pos_idx: Optional[int] = col.tbl.num_rowid_columns() - 1 if self.is_unstored_iter_col else None
        self.id = self._create_id()

    def set_iter_arg_ctx(self, iter_arg_ctx: RowBuilder.EvalCtx) -> None:
        self.iter_arg_ctx = iter_arg_ctx
        assert len(self.iter_arg_ctx.target_slot_idxs) == 1  # a single inline dict

    def _id_attrs(self) -> List[Tuple[str, Any]]:
        return super()._id_attrs() + [('tbl_id', self.col.tbl.id), ('col_id', self.col.id)]

    def __getattr__(self, name: str) -> Expr:
        from .column_property_ref import ColumnPropertyRef
        # resolve column properties
        if name == ColumnPropertyRef.Property.ERRORTYPE.name.lower() \
                or name == ColumnPropertyRef.Property.ERRORMSG.name.lower():
            if not (self.col.is_computed and self.col.is_stored) and not self.col.col_type.is_media_type():
                raise excs.Error(f'{name} only valid for a stored computed or media column: {self}')
            return ColumnPropertyRef(self, ColumnPropertyRef.Property[name.upper()])
        if name == ColumnPropertyRef.Property.FILEURL.name.lower() \
                or name == ColumnPropertyRef.Property.LOCALPATH.name.lower():
            if not self.col.col_type.is_media_type():
                raise excs.Error(f'{name} only valid for image/video/audio/document columns: {self}')
            if self.col.is_computed and not self.col.is_stored:
                raise excs.Error(f'{name} not valid for computed unstored columns: {self}')
            return ColumnPropertyRef(self, ColumnPropertyRef.Property[name.upper()])

        if self.col_type.is_json_type():
            from .json_path import JsonPath
            return JsonPath(self, [name])

        return super().__getattr__(name)

    def default_column_name(self) -> Optional[str]:
        return str(self)

    def _equals(self, other: ColumnRef) -> bool:
        return self.col == other.col

    def __str__(self) -> str:
        return self.col.name

    def sql_expr(self) -> Optional[sql.ClauseElement]:
        return self.col.sa_col

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        if not self.is_unstored_iter_col:
            assert data_row.has_val[self.slot_idx]
            return

        # if this is a new base row, we need to instantiate a new iterator
        if self.base_rowid != data_row.pk[:self.base_rowid_len]:
            row_builder.eval(data_row, self.iter_arg_ctx)
            iterator_args = data_row[self.iter_arg_ctx.target_slot_idxs[0]]
            self.iterator = self.col.tbl.iterator_cls(**iterator_args)
            self.base_rowid = data_row.pk[:self.base_rowid_len]
        self.iterator.set_pos(data_row.pk[self.pos_idx])
        res = next(self.iterator)
        data_row[self.slot_idx] = res[self.col.name]

    def _as_dict(self) -> Dict:
        tbl = self.col.tbl
        version = tbl.version if tbl.is_snapshot else None
        return {'tbl_id': str(tbl.id), 'tbl_version': version, 'col_id': self.col.id}

    @classmethod
    def _from_dict(cls, d: Dict, components: List[Expr]) -> Expr:
        tbl_id, version, col_id = UUID(d['tbl_id']), d['tbl_version'], d['col_id']
        tbl_version = catalog.Catalog.get().tbl_versions[(tbl_id, version)]
        assert col_id in tbl_version.cols_by_id
        col = tbl_version.cols_by_id[col_id]
        return cls(col)

